from __future__ import annotations
from angr.serializable import Serializable
from angr.protos import primitives_pb2
from .xref_types import XRefType


class XRef(Serializable):
    """
    XRef describes a reference to a MemoryData instance (if a MemoryData instance is available) or just an address.
    """

    __slots__ = (
        "block_addr",
        "dst",
        "ins_addr",
        "insn_op_idx",
        "insn_op_type",
        "memory_data",
        "stmt_idx",
        "type",
    )

    def __init__(
        self,
        ins_addr: int | None = None,
        block_addr: int | None = None,
        stmt_idx: int | None = None,
        insn_op_idx: int | None = None,
        memory_data: MemoryData | None = None,
        dst: int | None = None,
        xref_type=None,
    ):
        if dst is not None and not isinstance(dst, int):
            raise TypeError(f"XRefs must be pointing to a constant target. Target {dst!r} is not supported.")

        # src
        self.ins_addr: int | None = ins_addr
        self.insn_op_idx: int | None = insn_op_idx
        self.block_addr: int | None = block_addr
        self.stmt_idx: int | None = stmt_idx

        # dst
        self.memory_data = memory_data  # optional
        self.dst = dst
        self.type = xref_type

        if memory_data is not None and dst is None:
            self.dst = memory_data.addr

    @property
    def type_string(self):
        return XRefType.to_string(self.type)

    def __repr__(self):
        if self.dst is not None:
            dst_str = hex(self.dst) if isinstance(self.dst, int) else str(self.dst)
        elif self.memory_data is not None:
            dst_str = hex(self.memory_data.addr)
        else:
            dst_str = "unknown"
        ins_addr_str = f"{self.ins_addr:#x}" if self.ins_addr is not None else f"{self.block_addr:#x}[{self.stmt_idx}]"
        return f"<XRef {self.type_string}: {ins_addr_str}->{dst_str}>"

    def __eq__(self, other):
        return (
            type(other) is XRef
            and other.type == self.type
            and other.ins_addr == self.ins_addr
            and other.dst == self.dst
        )

    def __hash__(self):
        return hash((XRef, self.type, self.ins_addr, self.dst))

    @classmethod
    def _get_cmsg(cls):
        return primitives_pb2.CodeReference()

    def serialize_to_cmessage(self):
        # pylint:disable=no-member

        # delayed import
        from angr.engines.light import SpOffset  # pylint:disable=import-outside-toplevel

        cmsg = self._get_cmsg()
        if self.memory_data is not None:
            # determine target_type from memory_data.sort
            if self.memory_data.sort == MemoryDataSort.CodeReference:
                cmsg.target_type = primitives_pb2.CodeReference.CodeTarget
            else:
                cmsg.target_type = primitives_pb2.CodeReference.DataTarget

            cmsg.location = primitives_pb2.CodeReference.Internal
            cmsg.data_ea = self.memory_data.addr
        elif self.dst is not None:
            if isinstance(self.dst, SpOffset):
                cmsg.target_type = primitives_pb2.CodeReference.StackTarget
                cmsg.data_ea = self.dst.offset
            else:
                cmsg.data_ea = self.dst
        else:
            # Unknown... why?
            cmsg.data_ea = -1
        if self.insn_op_idx is None:
            cmsg.operand_idx = -1
        else:
            cmsg.operand_idx = self.insn_op_idx
        cmsg.ea = self.ins_addr
        cmsg.block_ea = self.block_addr
        cmsg.stmt_idx = self.stmt_idx
        cmsg.ref_type = self.type
        return cmsg

    @classmethod
    def parse_from_cmessage(cls, cmsg, bits=None, **kwargs):  # pylint:disable=arguments-differ
        # Note that we cannot recover _memory_data from cmsg

        # delayed import
        from angr.engines.light import SpOffset  # pylint:disable=import-outside-toplevel

        if not isinstance(bits, int):
            raise TypeError("bits must be provided.")

        if cmsg.target_type == primitives_pb2.CodeReference.StackTarget:  # pylint:disable=no-member
            # I think this code is dead - otherwise it would immediately hit the TypeError in the XRef constructor
            dst = SpOffset(bits, cmsg.data_ea, is_base=False)
        else:
            dst = int(cmsg.data_ea)

        return XRef(
            ins_addr=cmsg.ea,
            block_addr=cmsg.block_ea,
            stmt_idx=cmsg.stmt_idx,
            insn_op_idx=None if cmsg.operand_idx == -1 else cmsg.operand_idx,
            dst=dst,
            xref_type=cmsg.ref_type,
        )

    def copy(self):
        return XRef(
            ins_addr=self.ins_addr,
            block_addr=self.block_addr,
            stmt_idx=self.stmt_idx,
            insn_op_idx=self.insn_op_idx,
            memory_data=self.memory_data,
            dst=self.dst,
            xref_type=self.type,
        )


from angr.knowledge_plugins.cfg.memory_data import MemoryData, MemoryDataSort



from __future__ import annotations
from collections.abc import Callable

from angr import ailment
from angr.analyses import Analysis
from angr.analyses.decompiler.clinic import Clinic
from angr.analyses.forward_analysis.forward_analysis import ForwardAnalysisForClinic
from angr.errors import AngrAnalysisError
from angr.knowledge_plugins.functions.function import Function
from .engine import PurityEngineAIL, ResultType, StateType


class AILPurityAnalysis(Analysis, ForwardAnalysisForClinic[StateType]):
    """
    Determined how data sources are used in a given function.
    """

    def __init__(
        self,
        subject: str | int | Function | Clinic,
        memo: dict[Function, ResultType | None] | None = None,
        lifter: Callable[[Function], Clinic] | None = None,
    ):
        if memo is None:
            memo = {}
        self.memo = memo
        self._lifter = lifter or self._lift_default
        if isinstance(subject, (str, int)):
            subject = self.kb.functions[subject]
        if isinstance(subject, Function):
            subject = self._lifter(subject)
        self.engine = PurityEngineAIL(self.project, subject, self._recurse)
        self.result = ResultType()
        ForwardAnalysisForClinic.__init__(self, allow_merging=True, clinic=subject)

        memo[subject.function] = None
        self._analyze()
        memo[subject.function] = self.result

    def _lift_default(self, function: Function) -> Clinic:
        return self.project.analyses.Clinic(function)

    def _recurse(self, function: Function) -> ResultType:
        if function not in self.memo:
            self.project.analyses[AILPurityAnalysis].prep(kb=self.kb)(function, self.memo, self._lifter)
        r = self.memo[function]
        if r is None:
            raise AngrAnalysisError("Cannot do purity for recursive call tree")
        return r

    def _initial_abstract_state(self, node: ailment.Block) -> StateType:
        return self.engine.initial_state(node)

    def _step_node(self, node, state: StateType) -> tuple[bool | None, StateType]:
        state = state.copy()
        state.addr = (node.addr, node.idx)
        result = self.engine.process(state, block=node)
        self.result.update(result)
        return None, state

    def _handle_phi(self, node, succ, state, dst_vvar, src_vvar):
        state.vars[dst_vvar.varid] |= state.vars[src_vvar.varid]

    def _merge_states(self, node, *states) -> tuple[StateType, bool]:
        changed = False
        result = states[0]
        for state in states[1:]:
            for varid, val in state.vars.items():
                oval = result.vars[varid]
                if val - oval:
                    changed = True
                    oval |= val

        assert self.engine.clinic.graph is not None
        return result, changed

    def _compare_states(self, node, old_state, new_state) -> bool:
        return old_state.vars == new_state.vars



from __future__ import annotations
from typing import TYPE_CHECKING
import logging
from collections import defaultdict

import networkx

from angr.ailment.expression import Phi, VirtualVariable
from angr.ailment.statement import Assignment
from angr.knowledge_plugins.functions import Function
from angr.analyses import register_analysis
from .graph_rewriting import GraphRewritingAnalysis
from .dephication_base import DephicationBase

if TYPE_CHECKING:
    from angr import KnowledgeBase


l = logging.getLogger(name=__name__)


class GraphDephication(DephicationBase):  # pylint:disable=abstract-method
    """
    GraphDephication removes phi expressions from an AIL graph, essentially transforms a partial-SSA form of AIL graph
    to a normal AIL graph.
    """

    def __init__(
        self,
        func: Function | str,
        ail_graph,
        vvar_to_vvar_mapping: dict[int, int] | None = None,
        rewrite: bool = False,
        variable_kb: KnowledgeBase | None = None,
    ):
        """
        :param func:                            The subject of the analysis: a function, or a single basic block
        :param ail_graph:                       The AIL graph to transform.
        """

        self._graph = ail_graph

        super().__init__(func, vvar_to_vvar_mapping=vvar_to_vvar_mapping, rewrite=rewrite, variable_kb=variable_kb)

        self._analyze()

    def _collect_phi_assignments(self) -> dict[int, set[int]]:
        g = self._graph
        phi_to_src = defaultdict(set)

        for block in g:
            for stmt in block.statements:
                if isinstance(stmt, Assignment) and isinstance(stmt.dst, VirtualVariable) and isinstance(stmt.src, Phi):
                    for _, vvar in stmt.src.src_and_vvars:
                        if vvar is None:
                            l.debug("Invalid vvar None found in %r.src.src_and_vvars.", stmt)
                        else:
                            phi_to_src[stmt.dst.varid].add(vvar.varid)

        return phi_to_src

    def _rewrite_container(self) -> networkx.DiGraph:
        # replace all vvars with phi variables in the graph
        rewriter = GraphRewritingAnalysis(
            self.project, self._function, self._graph, self.vvar_to_vvar_mapping, variable_kb=self.variable_kb
        )
        return rewriter.out_graph


register_analysis(GraphDephication, "GraphDephication")



from __future__ import annotations
import claripy
from cle.backends.externs.simdata.io_file import io_file_data_for_arch

import angr

# Reference for implementation: glibc-2.25/libio/iofdopen.c


def mode_to_flag(mode):
    # TODO improve this: handle mode = strings
    if mode[-1] == b"b":  # lol who uses windows
        mode = mode[:-1]
    all_modes = {
        b"r": angr.storage.file.Flags.O_RDONLY,
        b"r+": angr.storage.file.Flags.O_RDWR,
        b"w": angr.storage.file.Flags.O_WRONLY | angr.storage.file.Flags.O_CREAT,
        b"w+": angr.storage.file.Flags.O_RDWR | angr.storage.file.Flags.O_CREAT,
        b"a": angr.storage.file.Flags.O_WRONLY | angr.storage.file.Flags.O_CREAT | angr.storage.file.Flags.O_APPEND,
        b"a+": angr.storage.file.Flags.O_RDWR | angr.storage.file.Flags.O_CREAT | angr.storage.file.Flags.O_APPEND,
    }
    if mode not in all_modes:
        raise angr.SimProcedureError(f"unsupported file open mode {mode}")

    return all_modes[mode]


def create_file(mode):
    # TODO improve this: handle mode = strings
    if mode[-1] == b"b":  # lol who uses windows
        mode = mode[:-1]
    all_modes = {
        b"r": False,
        b"r+": False,
        b"w": True,
        b"w+": True,
        b"a": True,
        b"a+": True,
    }
    if mode not in all_modes:
        raise angr.SimProcedureError(f"unsupported file open mode {mode}")

    return all_modes[mode]


class fdopen(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, fd_int, m_addr):
        # pylint:disable=unused-variable
        strlen = angr.SIM_PROCEDURES["libc"]["strlen"]

        m_strlen = self.inline_call(strlen, m_addr)
        m_expr = self.state.memory.load(m_addr, m_strlen.max_null_index, endness="Iend_BE")
        mode = self.state.solver.eval(m_expr, cast_to=bytes)

        # TODO: handle append and other mode subtleties

        fd_concr = self.state.posix.get_concrete_fd(fd_int, create_file=create_file(mode))
        if fd_concr not in self.state.posix.fd:
            # if file descriptor not found return NULL
            return 0
        # Allocate a FILE struct in heap
        malloc = angr.SIM_PROCEDURES["libc"]["malloc"]
        io_file_data = io_file_data_for_arch(self.state.arch)
        file_struct_ptr = self.inline_call(malloc, io_file_data["size"]).ret_expr

        # Write the fd
        fd_bvv = claripy.BVV(fd_concr, 4 * 8)  # int
        self.state.memory.store(file_struct_ptr + io_file_data["fd"], fd_bvv, endness=self.state.arch.memory_endness)

        if self.state.solver.is_true(fd_int == fd_concr):
            return file_struct_ptr
        null = claripy.BVV(0, self.state.arch.bits)
        return claripy.If(fd_int == fd_concr, file_struct_ptr, null)



from __future__ import annotations

import logging

import claripy

import angr

l = logging.getLogger(name=__name__)


class memccpy(angr.SimProcedure):
    # pylint:disable=arguments-differ, missing-class-docstring

    def run(self, dst_addr, src_addr, c_int, n):
        c = c_int[7:0]

        if self.state.solver.is_true(n == 0):
            return claripy.BVV(0, self.state.arch.bits)

        if not self.state.solver.symbolic(n):
            max_copy = self.state.solver.eval(n)
        else:
            max_copy = min(self.state.solver.max_int(n), self.state.libc.max_buffer_size)  # type: ignore

        if max_copy == 0:
            return claripy.BVV(0, self.state.arch.bits)

        l.debug("memccpy copying up to %d bytes until char found", max_copy)

        a, constraints, _i = self.state.memory.find(src_addr, c, max_copy, default=0)
        self.state.add_constraints(*constraints)

        c_pos = a - src_addr

        c_found = claripy.And(a != 0, claripy.ULT(c_pos, n)) if self.state.solver.symbolic(n) else a != 0

        copy_len = claripy.If(c_found, c_pos + 1, n)

        src_mem = self.state.memory.load(src_addr, max_copy, endness="Iend_BE")
        self.state.memory.store(dst_addr, src_mem, size=copy_len, endness="Iend_BE")

        return claripy.If(c_found, dst_addr + c_pos + 1, claripy.BVV(0, self.state.arch.bits))



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long,no-member
from __future__ import annotations

__package__ = __package__ or "tests"  # pylint:disable=redefined-builtin

import io
import logging
import os
import re
import sys
import unittest
from unittest import mock

from rich.console import Console
from rich.logging import RichHandler

import angr
from angr.__main__ import main
from angr.analyses.decompiler.decompiler import Decompiler
from angr.analyses.decompiler.utils import decompile_functions
from angr.analyses.disassembly import Disassembly

from .common import bin_location

test_location = os.path.join(bin_location, "tests")

ANSI_RE = re.compile(r"\x1b\[[0-9;]*[a-zA-Z]")


def run_cli(*args):
    with mock.patch("sys.argv", [sys.executable, *args]), mock.patch("sys.stdout", new=io.StringIO()) as fake_out:
        main()
        return fake_out.getvalue()


def run_cli_interactive(*args):
    """Run CLI simulating an interactive terminal. Returns (stdout, stderr) as plain text."""
    stderr_buf = io.StringIO()

    def mock_make_status():
        console = Console(file=stderr_buf, force_terminal=True, no_color=True, width=200)
        angr_logger = logging.getLogger("angr")
        angr_logger.handlers = [RichHandler(console=console, show_path=False, show_time=False)]
        angr_logger.propagate = False
        return console, True

    with (
        mock.patch("sys.argv", [sys.executable, *args]),
        mock.patch("sys.stdout", new=io.StringIO()) as fake_out,
        mock.patch("angr.__main__._make_status_console", mock_make_status),
    ):
        main()
        stderr_text = ANSI_RE.sub("", stderr_buf.getvalue())
        return fake_out.getvalue(), stderr_text


class TestCommandLineInterface(unittest.TestCase):
    def test_decompiling(self):
        bin_path = os.path.join(test_location, "x86_64", "decompiler", "sailr_motivating_example")
        f1 = "schedule_job"
        f2 = "main"

        # test a single function
        assert (
            run_cli("decompile", bin_path, "--functions", f1, "--no-colors")
            == decompile_functions(bin_path, [f1]) + "\n"
        )

        # test multiple functions
        assert (
            run_cli("decompile", bin_path, "--functions", f1, f2, "--no-colors")
            == decompile_functions(bin_path, [f1, f2]) + "\n"
        )

    def test_structuring(self):
        bin_path = os.path.join(test_location, "x86_64", "decompiler", "sailr_motivating_example")
        f1 = "schedule_job"

        dream_cli = run_cli("decompile", bin_path, "--no-casts", "--functions", f1, "--structurer", "dream")
        sailr_cli = run_cli("decompile", bin_path, "--functions", f1, "--structurer", "sailr")
        assert dream_cli.count("goto") == 0
        assert dream_cli != sailr_cli

    def test_base_addr_dec(self):
        bin_path = os.path.join(test_location, "x86_64", "decompiler", "sailr_motivating_example")
        f1 = "schedule_job"
        proj = angr.Project(bin_path, auto_load_libs=False)
        default_base_addr = proj.loader.main_object.min_addr
        f1_default_addr = proj.loader.find_symbol(f1).rebased_addr
        f1_offset = f1_default_addr - default_base_addr

        # function resolving is based on symbol
        sym_based_dec = run_cli("decompile", bin_path, "--functions", f1, "--preset", "full", "--no-colors")
        # function resolving is based on the address (with default angr loading)
        base_addr_dec = run_cli(
            "decompile", bin_path, "--functions", hex(f1_default_addr), "--preset", "full", "--no-colors"
        )
        # function resolving is based on the address (with base address specified)
        offset_dec = run_cli(
            "decompile",
            bin_path,
            "--base-addr",
            "0x0",
            "--functions",
            hex(f1_offset),
            "--preset",
            "full",
            "--no-colors",
        )

        # since the externs can be unpredictable, we only check the function name down
        sym_based_dec = re.sub(r"extern .*;", "", sym_based_dec).lstrip("\n")
        base_addr_dec = re.sub(r"extern .*;", "", base_addr_dec).lstrip("\n")
        offset_dec = re.sub(r"extern .*;", "", offset_dec).lstrip("\n")

        # we must also normalize label names (since they are based on the address)
        sym_based_dec = re.sub(r"LABEL_[0-9a-fA-F]+", "LABEL_A", sym_based_dec)
        base_addr_dec = re.sub(r"LABEL_[0-9a-fA-F]+", "LABEL_A", base_addr_dec)
        offset_dec = re.sub(r"LABEL_[0-9a-fA-F]+", "LABEL_A", offset_dec)

        assert sym_based_dec == base_addr_dec == offset_dec

    def test_disassembly(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        disasm = run_cli("disassemble", bin_path)
        funcs = {
            "_init",
            "_start",
            "call_gmon_start",
            "__do_global_dtors_aux",
            "frame_dummy",
            "authenticate",
            "accepted",
            "rejected",
            "main",
            "__libc_csu_init",
            "__libc_csu_fini",
            "__do_global_ctors_aux",
            "_fini",
        }
        substrs = [f"{f}:" for f in funcs]
        substrs += "40071d  55              push    rbp"

        for s in substrs:
            assert s in disasm

    def test_syntax_highlighting_no_colors_flag(self):
        bin_path = os.path.join(test_location, "x86_64", "decompiler", "sailr_motivating_example")
        f1 = "main"

        no_colors_output = run_cli("decompile", bin_path, "--functions", f1, "--no-colors")
        expected_output = decompile_functions(bin_path, [f1]) + "\n"

        # it should maintain that no ANSI color codes are present
        assert no_colors_output == expected_output

    def test_aliases(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        dec_output = run_cli("dec", bin_path, "--functions", "main", "--no-colors")
        assert "main" in dec_output

        dis_output = run_cli("dis", bin_path, "--functions", "main")
        assert "main:" in dis_output

    def test_disassemble_specific_functions(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        output = run_cli("disassemble", bin_path, "--functions", "main")
        assert "main:" in output
        assert "authenticate:" not in output

    def test_disassemble_function_by_address(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        proj = angr.Project(bin_path, auto_load_libs=False)
        main_addr = proj.loader.find_symbol("main").rebased_addr
        output = run_cli("disassemble", bin_path, "--functions", hex(main_addr))
        assert "main:" in output

    def test_decompile_interactive_summary(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        stdout, stderr = run_cli_interactive(
            "decompile", bin_path, "--functions", "main", "authenticate", "--no-colors"
        )
        # Both functions should appear in stdout
        assert "main" in stdout
        # Summary on stderr for multiple functions
        assert "Decompiled" in stderr

    def test_disassemble_interactive_summary(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        stdout, stderr = run_cli_interactive("disassemble", bin_path, "--functions", "main", "authenticate")
        assert "main:" in stdout
        assert "authenticate:" in stdout
        assert "Disassembled" in stderr

    def test_decompile_missing_function_raises(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        with self.assertRaises(ValueError):
            run_cli("decompile", bin_path, "--functions", "nonexistent_func", "--no-colors")

    def test_decompile_catch_exceptions_missing_function(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        stdout, stderr = run_cli_interactive(
            "decompile", bin_path, "--functions", "nonexistent_func", "--catch-exceptions", "--no-colors"
        )
        assert stdout == ""
        assert "not found" in stderr.lower()

    def test_disassemble_missing_function(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        # parse_function_args logs error and yields nothing -> "No disassemblable functions found"
        stdout, stderr = run_cli_interactive("disassemble", bin_path, "--functions", "nonexistent_func")
        assert stdout == ""
        assert "no disassemblable functions found" in stderr.lower()

    def test_decompile_no_decompilable_functions(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        # _init is typically a stub that gets filtered out
        _stdout, stderr = run_cli_interactive(
            "decompile", bin_path, "--functions", "nonexistent_func", "--catch-exceptions", "--no-colors"
        )
        assert "no decompilable functions found" in stderr.lower() or "not found" in stderr.lower()

    def test_decompile_cca(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        output = run_cli("decompile", bin_path, "--functions", "main", "--cca", "--no-colors")
        assert "main" in output

    def test_decompile_syntax_highlighting(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        # Force ansi_color_enabled=True to exercise the Syntax highlighting path.
        # Note: Rich's Console strips ANSI when stdout is a StringIO (not a real terminal),
        # so we verify the path ran by checking the output content.
        with mock.patch("angr.__main__.ansi_color_enabled", True):
            stdout, _ = run_cli_interactive("decompile", bin_path, "--functions", "main")
        assert stdout  # non-empty
        assert "main" in ANSI_RE.sub("", stdout)

    def test_decompile_syntax_highlighting_multiple(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        with mock.patch("angr.__main__.ansi_color_enabled", True):
            stdout, stderr = run_cli_interactive("decompile", bin_path, "--functions", "main", "authenticate")
        plain = ANSI_RE.sub("", stdout)
        assert "main" in plain
        assert "authenticate" in plain
        assert "Decompiled" in stderr

    def test_disassemble_base_addr(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        output = run_cli("disassemble", bin_path, "--base-addr", "0x0", "--functions", "main")
        assert "main:" in output

    def test_disassemble_missing_address(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        # Address passes numeric regex but doesn't match any function
        stdout, stderr = run_cli_interactive("disassemble", bin_path, "--functions", "0xdeadbeef")
        assert stdout == ""
        assert "no disassemblable functions found" in stderr.lower()

    def test_decompile_catch_exceptions_noninteractive(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        # Non-interactive catch-exceptions hits the log.warning path
        output = run_cli("decompile", bin_path, "--functions", "nonexistent_func", "--catch-exceptions", "--no-colors")
        assert output == ""

    def test_decompile_catch_exceptions_error(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        with mock.patch.object(Decompiler, "_decompile", side_effect=RuntimeError("mock failure")):
            stdout, stderr = run_cli_interactive(
                "decompile", bin_path, "--functions", "main", "authenticate", "--catch-exceptions", "--no-colors"
            )
        assert stdout == ""
        assert "error decompiling" in stderr.lower()
        assert "2 error" in stderr.lower()

    def test_disassemble_catch_exceptions_error(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        with mock.patch.object(Disassembly, "render", side_effect=RuntimeError("mock failure")):
            stdout, stderr = run_cli_interactive(
                "disassemble", bin_path, "--functions", "main", "authenticate", "--catch-exceptions"
            )
        assert stdout == ""
        assert "error disassembling" in stderr.lower()
        assert "2 error" in stderr.lower()

    def test_decompile_no_output(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        # Simulate decompilation succeeding but producing no codegen output
        with mock.patch.object(Decompiler, "_decompile"):
            stdout, stderr = run_cli_interactive(
                "decompile", bin_path, "--functions", "main", "--catch-exceptions", "--no-colors"
            )
        assert stdout == ""
        assert "no output" in stderr.lower() or "error decompiling" in stderr.lower()

    def test_disassemble_catch_exceptions_noninteractive(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        with mock.patch.object(Disassembly, "render", side_effect=RuntimeError("mock failure")):
            output = run_cli("disassemble", bin_path, "--functions", "main", "--catch-exceptions")
        assert output == ""

    def test_decompile_no_functions_noninteractive(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        # _init is a stub that gets filtered; requesting only it should yield nothing
        output = run_cli("decompile", bin_path, "--functions", "nonexistent_func", "--catch-exceptions", "--no-colors")
        assert output == ""

    def test_disassemble_no_functions_noninteractive(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        output = run_cli("disassemble", bin_path, "--functions", "nonexistent_func")
        assert output == ""

    def test_decompile_syntax_highlighting_catch_exceptions(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        # Highlighted mode with catch-exceptions: error should go to stderr, not stdout
        with (
            mock.patch("angr.__main__.ansi_color_enabled", True),
            mock.patch.object(Decompiler, "_decompile", side_effect=RuntimeError("mock failure")),
        ):
            stdout, stderr = run_cli_interactive("decompile", bin_path, "--functions", "main", "--catch-exceptions")
        assert stdout == ""
        assert "error decompiling" in stderr.lower()

    def test_blob_requires_required_args(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        with (
            mock.patch("sys.argv", [sys.executable, "disassemble", bin_path, "--blob"]),
            mock.patch("sys.stderr", new=io.StringIO()) as fake_err,
            self.assertRaises(SystemExit) as exc,
        ):
            main()

        assert exc.exception.code == 2
        assert "--blob requires --arch" in fake_err.getvalue()

    def test_disassemble_blob_loader_options(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        fake_proj = mock.Mock()
        fake_proj.analyses.CFG.return_value = None
        fake_proj.kb.functions = {}

        with mock.patch("angr.__main__.angr.Project", return_value=fake_proj) as mock_project:
            output = run_cli(
                "disassemble",
                bin_path,
                "--blob",
                "--arch",
                "AMD64",
                "--base-addr",
                "0x400000",
                "--entry-point",
                "0x400000",
            )

        assert output == ""
        _, kwargs = mock_project.call_args
        assert kwargs["auto_load_libs"] is False
        assert kwargs["main_opts"] == {
            "backend": "blob",
            "arch": "AMD64",
            "base_addr": 0x400000,
            "entry_point": 0x400000,
        }

    def test_decompile_blob_loader_options(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")

        fake_cfg = mock.Mock()
        fake_cfg.kb.functions = {}
        fake_cfg.functions = {}

        fake_proj = mock.Mock()
        fake_proj.analyses.CFG.return_value = fake_cfg

        with mock.patch("angr.__main__.angr.Project", return_value=fake_proj) as mock_project:
            output = run_cli(
                "decompile",
                bin_path,
                "--blob",
                "--arch",
                "AMD64",
                "--base-addr",
                "0x400000",
                "--entry-point",
                "0x400000",
                "--no-colors",
            )

        assert output == ""
        _, kwargs = mock_project.call_args
        assert kwargs["auto_load_libs"] is False
        assert kwargs["main_opts"] == {
            "backend": "blob",
            "arch": "AMD64",
            "base_addr": 0x400000,
            "entry_point": 0x400000,
        }


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
from collections import defaultdict, OrderedDict
from typing import Any
from collections.abc import Callable
from collections.abc import Generator
import operator
import logging

import angr.ailment as ailment
import claripy
import networkx


from angr.utils.graph import GraphUtils
from angr.utils.graph import dominates, inverted_idoms
from angr.utils.ail import is_head_controlled_loop_block
from angr.block import Block, BlockNode
from angr.errors import AngrRuntimeError
from .peephole_optimizations import InvertNegatedLogicalConjunctionsAndDisjunctions, RemoveRedundantNots
from .structuring.structurer_nodes import (
    MultiNode,
    EmptyBlockNotice,
    SequenceNode,
    CodeNode,
    SwitchCaseNode,
    BreakNode,
    ConditionalBreakNode,
    LoopNode,
    ConditionNode,
    ContinueNode,
    CascadingConditionNode,
    IncompleteSwitchCaseNode,
)
from .graph_region import GraphRegion
from .utils import peephole_optimize_expr

l = logging.getLogger(__name__)


_UNIFIABLE_COMPARISONS = {
    "__ne__",
    "__gt__",
    "__ge__",
    "UGT",
    "UGE",
    "SGT",
    "SGE",
}


_INVERSE_OPERATIONS = {
    "__eq__": "__ne__",
    "__ne__": "__eq__",
    "__gt__": "__le__",
    "__lt__": "__ge__",
    "__ge__": "__lt__",
    "__le__": "__gt__",
    "ULT": "UGE",
    "UGE": "ULT",
    "UGT": "ULE",
    "ULE": "UGT",
    "SLT": "SGE",
    "SGE": "SLT",
    "SLE": "SGT",
    "SGT": "SLE",
}


class AILExprIdAnnotation(claripy.Annotation):
    """
    An annotation that we use to annotate BVVs so that they are differentiable between other BVVs with the same value
    and size.
    """

    @property
    def eliminatable(self):
        return True

    @property
    def relocateable(self):
        return False


#
# Util methods and mapping used during AIL AST to claripy AST conversion
#


def _op_with_unified_size(op, conv: Callable, operand0, operand1, ins_addr: int):
    # ensure operand1 is of the same size as operand0
    if isinstance(operand1, ailment.Expr.Const):
        # amazing - we do the easy thing here
        return op(conv(operand0, nobool=True, ins_addr=ins_addr), operand1.value)
    if operand1.bits == operand0.bits:
        return op(conv(operand0, nobool=True, ins_addr=ins_addr), conv(operand1, ins_addr=ins_addr))
    # extension is required
    assert operand1.bits < operand0.bits
    operand1 = ailment.Expr.Convert(None, operand1.bits, operand0.bits, False, operand1)
    return op(conv(operand0, nobool=True, ins_addr=ins_addr), conv(operand1, nobool=True, ins_addr=ins_addr))


def _dummy_bvs(condition, condition_mapping, name_suffix="", must_bool=False):
    if must_bool:
        var = claripy.BoolS(f"ailexpr_{condition!r}{name_suffix}", explicit_name=True)
    else:
        var = claripy.BVS(f"ailexpr_{condition!r}{name_suffix}", condition.bits, explicit_name=True)
    condition_mapping[var.args[0]] = condition
    return var


def _dummy_bools(condition, condition_mapping, name_suffix=""):
    var = claripy.BoolS(f"ailexpr_{condition!r}{name_suffix}", explicit_name=True)
    condition_mapping[var.args[0]] = condition
    return var


_ail2claripy_op_mapping = {
    "LogicalAnd": lambda expr, conv, _, ia: claripy.And(
        conv(expr.operands[0], ins_addr=ia), conv(expr.operands[1], ins_addr=ia)
    ),
    "LogicalOr": lambda expr, conv, _, ia: claripy.Or(
        conv(expr.operands[0], ins_addr=ia), conv(expr.operands[1], ins_addr=ia)
    ),
    "CmpEQ": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) == conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CmpNE": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) != conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CmpLE": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) <= conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CmpLE (signed)": lambda expr, conv, _, ia: claripy.SLE(
        conv(expr.operands[0], nobool=True, ins_addr=ia), conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CmpLT": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) < conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CmpLT (signed)": lambda expr, conv, _, ia: claripy.SLT(
        conv(expr.operands[0], nobool=True, ins_addr=ia), conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CmpGE": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) >= conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CmpGE (signed)": lambda expr, conv, _, ia: claripy.SGE(
        conv(expr.operands[0], nobool=True, ins_addr=ia), conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CmpGT": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) > conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CmpGT (signed)": lambda expr, conv, _, ia: claripy.SGT(
        conv(expr.operands[0], nobool=True, ins_addr=ia), conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CasCmpEQ": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) == conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CasCmpNE": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) != conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CasCmpLE": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) <= conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CasCmpLE (signed)": lambda expr, conv, _, ia: claripy.SLE(
        conv(expr.operands[0], nobool=True, ins_addr=ia), conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CasCmpLT": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) < conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CasCmpLT (signed)": lambda expr, conv, _, ia: claripy.SLT(
        conv(expr.operands[0], nobool=True, ins_addr=ia), conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CasCmpGE": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) >= conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CasCmpGE (signed)": lambda expr, conv, _, ia: claripy.SGE(
        conv(expr.operands[0], nobool=True, ins_addr=ia), conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CasCmpGT": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) > conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "CasCmpGT (signed)": lambda expr, conv, _, ia: claripy.SGT(
        conv(expr.operands[0], nobool=True, ins_addr=ia), conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "Add": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) + conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "Sub": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) - conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "Mul": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) * conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "Div": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) / conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "Mod": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) % conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "Not": lambda expr, conv, _, ia: claripy.Not(conv(expr.operand, ins_addr=ia)),
    "Neg": lambda expr, conv, _, ia: -conv(expr.operand, ins_addr=ia),
    "BitwiseNeg": lambda expr, conv, _, ia: ~conv(expr.operand, ins_addr=ia),
    "Xor": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) ^ conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "And": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) & conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "Or": lambda expr, conv, _, ia: (
        conv(expr.operands[0], nobool=True, ins_addr=ia) | conv(expr.operands[1], nobool=True, ins_addr=ia)
    ),
    "Shr": lambda expr, conv, _, ia: _op_with_unified_size(claripy.LShR, conv, expr.operands[0], expr.operands[1], ia),
    "Shl": lambda expr, conv, _, ia: _op_with_unified_size(
        operator.lshift, conv, expr.operands[0], expr.operands[1], ia
    ),
    "Sar": lambda expr, conv, _, ia: _op_with_unified_size(
        operator.rshift, conv, expr.operands[0], expr.operands[1], ia
    ),
    "Concat": lambda expr, conv, _, ia: claripy.Concat(*[conv(operand, ins_addr=ia) for operand in expr.operands]),
    # There are no corresponding claripy operations for the following operations
    "CmpF": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "Mull": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "Mull (signed)": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "Reinterpret": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "Rol": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "Ror": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "LogicalXor": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "Carry": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "SCarry": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "SBorrow": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "ExpCmpNE": lambda expr, _, m, *args: _dummy_bools(expr, m),
    "CmpORD": lambda expr, _, m, *args: _dummy_bvs(expr, m),  # in case CmpORDRewriter fails
    "CmpEQV": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "GetMSBs": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "ShlNV": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "ShrNV": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "InterleaveLOV": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    "InterleaveHIV": lambda expr, _, m, *args: _dummy_bvs(expr, m),
    # catch-all
    "_DUMMY_": lambda expr, _, m, *args: _dummy_bvs(expr, m),
}

#
# The ConditionProcessor class
#


class ConditionProcessor:
    """
    Convert between claripy AST and AIL expressions. Also calculates reaching conditions of all nodes on a graph.
    """

    def __init__(self, arch, condition_mapping=None):
        self.arch = arch
        self._condition_mapping: dict[str, Any] = {} if condition_mapping is None else condition_mapping
        self.jump_table_conds: dict[int, set] = defaultdict(set)
        self.reaching_conditions = {}
        self.guarding_conditions = {}
        self._ast2annotations = {}

        self._peephole_expr_optimizations = [
            cls(None, None, None) for cls in [InvertNegatedLogicalConjunctionsAndDisjunctions, RemoveRedundantNots]
        ]

    def clear(self):
        self._condition_mapping = {}
        self.jump_table_conds = defaultdict(set)
        self.reaching_conditions = {}
        self.guarding_conditions = {}
        self._ast2annotations = {}

    def have_opposite_edge_conditions(self, graph: networkx.DiGraph, src, dst0, dst1) -> bool:
        """
        Check if the edge conditions of two edges (src, dst0) and (src, dst1) are opposite to each other. Try to avoid
        condition translation if possible.
        """

        if isinstance(src, SequenceNode) and src.nodes and isinstance(src.nodes[-1], ConditionNode):
            cond_node = src.nodes[-1]
            if (
                isinstance(cond_node.true_node, ailment.Block)
                and isinstance(cond_node.false_node, ailment.Block)
                and cond_node.true_node.statements
                and cond_node.false_node.statements
            ):
                last_stmt_true = self.get_last_statement(cond_node.true_node)
                last_stmt_false = self.get_last_statement(cond_node.false_node)
                if (
                    isinstance(last_stmt_true, ailment.Stmt.Jump)
                    and isinstance(last_stmt_false, ailment.Stmt.Jump)
                    and isinstance(last_stmt_true.target, ailment.Expr.Const)
                    and isinstance(last_stmt_false.target, ailment.Expr.Const)
                ):
                    return {last_stmt_true.target.value, last_stmt_false.target.value} == {dst0.addr, dst1.addr}

        if src in graph and graph.out_degree[src] == 2 and graph.has_edge(src, dst0) and graph.has_edge(src, dst1):
            # sometimes the last statement is the conditional jump. sometimes it's the first statement of the block
            if isinstance(src, ailment.Block) and src.statements and is_head_controlled_loop_block(src):
                last_stmt = next(
                    iter(stmt for stmt in src.statements[:-1] if isinstance(stmt, ailment.Stmt.ConditionalJump)), None
                )
                assert last_stmt is not None
            else:
                try:
                    last_stmt = self.get_last_statement(src)
                except EmptyBlockNotice:
                    last_stmt = None

            if isinstance(last_stmt, ailment.Stmt.ConditionalJump):
                return True

        # fallback
        edge_cond_left = self.recover_edge_condition(graph, src, dst0)
        edge_cond_right = self.recover_edge_condition(graph, src, dst1)
        cond = claripy.Not(edge_cond_left) == edge_cond_right
        # call claripy.simplify() just in case there are annotations
        return claripy.is_true(claripy.simplify(cond))  # type: ignore

    def recover_edge_condition(self, graph: networkx.DiGraph, src, dst):
        def _check_condnode_and_get_condition(cond_node: ConditionNode) -> claripy.ast.Bool | None:
            for cond_block, negate in [(cond_node.true_node, False), (cond_node.false_node, True)]:
                if isinstance(cond_block, ailment.Block) and cond_block.statements:
                    last_stmt = self.get_last_statement(cond_block)
                    if (
                        isinstance(last_stmt, ailment.Stmt.Jump)
                        and isinstance(last_stmt.target, ailment.Expr.Const)
                        and last_stmt.target.value == dst.addr
                    ):
                        return claripy.Not(cond_node.condition) if negate else cond_node.condition
            return None

        if isinstance(src, SequenceNode) and src.nodes and isinstance(src.nodes[-1], ConditionNode):
            predicate = _check_condnode_and_get_condition(src.nodes[-1])
            if predicate is not None:
                return predicate
        if isinstance(src, ConditionNode):
            predicate = _check_condnode_and_get_condition(src)
            if predicate is not None:
                return predicate

        edge = src, dst
        edge_data = graph.get_edge_data(*edge)
        edge_type = edge_data.get("type", "transition") if edge_data is not None else "transition"
        try:
            predicate = self._extract_predicate(src, dst, edge_type)
        except EmptyBlockNotice:
            # catch empty block notice - although this should not really happen
            predicate = claripy.true()
        return predicate

    def recover_edge_conditions(self, region, graph=None) -> dict:
        edge_conditions = {}
        # traverse the graph to recover the condition for each edge
        graph = graph or region.graph
        for src in graph.nodes():
            nodes = list(graph[src])
            if len(nodes) >= 1:
                for dst in nodes:
                    predicate = self.recover_edge_condition(graph, src, dst)
                    edge_conditions[(src, dst)] = predicate

        return edge_conditions

    def recover_reaching_conditions(
        self,
        region,
        graph=None,
        with_successors=False,
        case_entry_to_switch_head: dict[int, int] | None = None,
        simplify_conditions: bool = True,
    ):
        """
        Recover the reaching conditions for each block in an acyclic graph. Note that we assume the graph that's passed
        in is acyclic.
        """

        def _strictly_postdominates(inv_idoms, node_a, node_b):
            """
            Does node A strictly post-dominate node B on the graph?
            """
            return dominates(inv_idoms, node_a, node_b)

        edge_conditions = self.recover_edge_conditions(region, graph=graph)

        if graph:
            _g = graph
            head = next(node for node in graph.nodes if graph.in_degree(node) == 0)
        else:
            if with_successors and region.graph_with_successors is not None:
                _g = region.graph_with_successors
            else:
                _g = region.graph
            head = region.head

        # special handling for jump table entries - do not allow crossing between cases
        if case_entry_to_switch_head:
            _g = self._remove_crossing_edges_between_cases(_g, case_entry_to_switch_head)

        inverted_graph, idoms = inverted_idoms(_g)

        reaching_conditions = {}
        # recover the reaching condition for each node
        sorted_nodes = GraphUtils.quasi_topological_sort_nodes(_g)
        terminating_nodes = []
        for node in sorted_nodes:
            # create special conditions for all nodes that are jump table entries
            if case_entry_to_switch_head and node.addr in case_entry_to_switch_head:
                jump_target_var = self.create_jump_target_var(case_entry_to_switch_head[node.addr])
                cond = jump_target_var == claripy.BVV(node.addr, self.arch.bits)
                reaching_conditions[node] = cond
                self.jump_table_conds[case_entry_to_switch_head[node.addr]].add(cond)
                continue

            preds = _g.predecessors(node)
            reaching_condition = None

            out_degree = _g.out_degree(node)
            if out_degree == 0:
                terminating_nodes.append(node)

            if node is head:
                # the head is always reachable
                reaching_condition = claripy.true()
            elif idoms is not None and _strictly_postdominates(idoms, node, head):
                # the node that post dominates the head is always reachable
                reaching_conditions[node] = claripy.true()
            else:
                for pred in preds:
                    edge = (pred, node)
                    pred_condition = reaching_conditions.get(pred, claripy.true())
                    edge_condition = edge_conditions.get(edge, claripy.true())

                    if reaching_condition is None:
                        reaching_condition = claripy.And(pred_condition, edge_condition)
                    else:
                        reaching_condition = claripy.Or(claripy.And(pred_condition, edge_condition), reaching_condition)

            if reaching_condition is not None:
                reaching_conditions[node] = (
                    self.simplify_condition(reaching_condition) if simplify_conditions else reaching_condition
                )

        # My hypothesis: for nodes where two paths come together *and* those that cannot be further structured into
        # another if-else construct (we take the short-cut by testing if the operator is an "Or" after running our
        # condition simplifiers previously), we are better off using their "guarding conditions" instead of their
        # reaching conditions for if-else. see my super long chatlog with rhelmot on 5/14/2021.
        guarding_conditions = {}
        for the_node in sorted_nodes:
            preds = list(_g.predecessors(the_node))
            if len(preds) != 2:
                continue
            # generate a graph slice that goes from the region head to this node
            slice_nodes = list(networkx.dfs_tree(inverted_graph, the_node))
            subgraph = networkx.subgraph(_g, slice_nodes)
            # figure out which paths cause the divergence from this node
            nodes_do_not_reach_the_node = set()
            for node_ in subgraph:
                if node_ is the_node:
                    continue
                for succ in _g.successors(node_):
                    if not networkx.has_path(_g, succ, the_node):
                        nodes_do_not_reach_the_node.add(succ)

            diverging_conditions = []

            for node_ in nodes_do_not_reach_the_node:
                preds_ = list(_g.predecessors(node_))
                for pred_ in preds_:
                    if pred_ in nodes_do_not_reach_the_node:
                        continue
                    # this predecessor is the diverging node!
                    edge_ = pred_, node_
                    edge_condition = edge_conditions.get(edge_, None)
                    if edge_condition is not None:
                        diverging_conditions.append(edge_condition)

            if diverging_conditions:
                # the negation of the union of diverging conditions is the guarding condition for this node
                cond = claripy.Or(*map(claripy.Not, diverging_conditions))  # pylint:disable=bad-builtin
                guarding_conditions[the_node] = cond

        self.reaching_conditions = reaching_conditions
        self.guarding_conditions = guarding_conditions

    def remove_claripy_bool_asts(self, node, memo=None):
        # Convert claripy Bool ASTs to AIL expressions

        if memo is None:
            memo = {}

        if isinstance(node, SequenceNode):
            new_nodes = []
            for n in node.nodes:
                new_node = self.remove_claripy_bool_asts(n, memo=memo)
                new_nodes.append(new_node)
            return SequenceNode(node.addr, new_nodes)

        if isinstance(node, MultiNode):
            new_nodes = []
            for n in node.nodes:
                new_node = self.remove_claripy_bool_asts(n, memo=memo)
                new_nodes.append(new_node)
            return MultiNode(nodes=new_nodes)

        if isinstance(node, CodeNode):
            return CodeNode(
                self.remove_claripy_bool_asts(node.node, memo=memo),
                (
                    None
                    if node.reaching_condition is None
                    else self.convert_claripy_bool_ast(node.reaching_condition, memo=memo)
                ),
            )

        if isinstance(node, ConditionalBreakNode):
            return ConditionalBreakNode(
                node.addr,
                self.convert_claripy_bool_ast(node.condition, memo=memo),
                node.target,
            )

        if isinstance(node, ConditionNode):
            return ConditionNode(
                node.addr,
                (
                    None
                    if node.reaching_condition is None
                    else self.convert_claripy_bool_ast(node.reaching_condition, memo=memo)
                ),
                self.convert_claripy_bool_ast(node.condition, memo=memo),
                self.remove_claripy_bool_asts(node.true_node, memo=memo),
                self.remove_claripy_bool_asts(node.false_node, memo=memo),
            )

        if isinstance(node, CascadingConditionNode):
            cond_and_nodes = []
            for cond, child_node in node.condition_and_nodes:
                cond_and_nodes.append(
                    (
                        self.convert_claripy_bool_ast(cond, memo=memo),
                        self.remove_claripy_bool_asts(child_node, memo=memo),
                    )
                )
            else_node = None if node.else_node is None else self.remove_claripy_bool_asts(node.else_node, memo=memo)
            return CascadingConditionNode(
                node.addr,
                cond_and_nodes,
                else_node=else_node,
            )

        if isinstance(node, LoopNode):
            result = node.copy()
            result.condition = (
                self.convert_claripy_bool_ast(node.condition, memo=memo) if node.condition is not None else None
            )
            result.sequence_node = self.remove_claripy_bool_asts(node.sequence_node, memo=memo)
            return result

        if isinstance(node, SwitchCaseNode):
            return SwitchCaseNode(
                self.convert_claripy_bool_ast(node.switch_expr, memo=memo),
                OrderedDict(
                    (idx, self.remove_claripy_bool_asts(case_node, memo=memo)) for idx, case_node in node.cases.items()
                ),
                self.remove_claripy_bool_asts(node.default_node, memo=memo),
                addr=node.addr,
            )

        if isinstance(node, IncompleteSwitchCaseNode):
            return IncompleteSwitchCaseNode(
                node.addr,
                self.remove_claripy_bool_asts(node.head, memo=memo),
                [self.remove_claripy_bool_asts(case, memo=memo) for case in node.cases],
            )

        return node

    @classmethod
    def get_last_statement(cls, block):
        """
        This is the buggy version of get_last_statements, because, you know, there can always be more than one last
        statement due to the existence of branching statements (like, If-then-else). All methods using
        get_last_statement() should switch to get_last_statements() and properly handle multiple last statements.
        """
        if type(block) is SequenceNode:
            if block.nodes:
                return cls.get_last_statement(block.nodes[-1])
            raise EmptyBlockNotice
        if type(block) is CodeNode:
            return cls.get_last_statement(block.node)
        if type(block) is ailment.Block:
            if not block.statements:
                raise EmptyBlockNotice
            return block.statements[-1]
        if type(block) is Block:
            raise NotImplementedError
        if type(block) is BlockNode:
            raise NotImplementedError
        if type(block) is MultiNode:
            # get the last node
            for the_block in reversed(block.nodes):
                try:
                    return cls.get_last_statement(the_block)
                except EmptyBlockNotice:
                    continue
            raise EmptyBlockNotice
        if type(block) is LoopNode:
            return cls.get_last_statement(block.sequence_node)
        if type(block) is ConditionalBreakNode:
            return None
        if type(block) is ConditionNode:
            s = None
            if block.true_node:
                try:
                    s = cls.get_last_statement(block.true_node)
                except EmptyBlockNotice:
                    s = None
            if s is None and block.false_node:
                s = cls.get_last_statement(block.false_node)
            return s
        if type(block) is CascadingConditionNode:
            s = None
            if block.else_node is not None:
                s = cls.get_last_statement(block.else_node)
            else:
                for _, node in reversed(block.condition_and_nodes):
                    s = cls.get_last_statement(node)
                    if s is not None:
                        break
            return s
        if type(block) is BreakNode:
            return None
        if type(block) is ContinueNode:
            return None
        if type(block) is SwitchCaseNode:
            return None
        if type(block) is IncompleteSwitchCaseNode:
            return None
        if type(block) is GraphRegion:
            # normally this should not happen. however, we have test cases that trigger this case.
            return None

        raise NotImplementedError

    @classmethod
    def get_last_statements(
        cls, block
    ) -> list[ailment.Stmt.Statement | ConditionalBreakNode | BreakNode | ContinueNode | None]:
        if type(block) is SequenceNode:
            for last_node in reversed(block.nodes):
                try:
                    return cls.get_last_statements(last_node)
                except EmptyBlockNotice:
                    # the node is empty. try the next one
                    continue

            raise EmptyBlockNotice

        if type(block) is CodeNode:
            return cls.get_last_statements(block.node)
        if type(block) is ailment.Block:
            if not block.statements:
                raise EmptyBlockNotice
            return [block.statements[-1]]
        if type(block) is Block:
            raise NotImplementedError
        if type(block) is BlockNode:
            raise NotImplementedError
        if type(block) is MultiNode:
            # get the last node
            for the_block in reversed(block.nodes):
                try:
                    return cls.get_last_statements(the_block)
                except EmptyBlockNotice:
                    continue
            raise EmptyBlockNotice
        if type(block) is LoopNode:
            if block.sequence_node is None:
                raise EmptyBlockNotice
            return cls.get_last_statements(block.sequence_node)
        if type(block) is ConditionalBreakNode:
            return [block]
        if type(block) is ConditionNode:
            s = []
            if block.true_node:
                try:
                    last_stmts = cls.get_last_statements(block.true_node)
                    s.extend(last_stmts)
                except EmptyBlockNotice:
                    pass
            else:
                s.append(None)
            if block.false_node:
                last_stmts = cls.get_last_statements(block.false_node)
                s.extend(last_stmts)
            else:
                s.append(None)
            return s
        if type(block) is CascadingConditionNode:
            s = []
            if block.else_node is not None:
                try:
                    last_stmts = cls.get_last_statements(block.else_node)
                    s.extend(last_stmts)
                except EmptyBlockNotice:
                    pass
            else:
                s.append(None)
            for _, node in block.condition_and_nodes:
                last_stmts = cls.get_last_statements(node)
                s.extend(last_stmts)
            return s
        if type(block) is BreakNode:
            return [block]
        if type(block) is ContinueNode:
            return [block]
        if type(block) is SwitchCaseNode:
            s = []
            for case in block.cases.values():
                s.extend(cls.get_last_statements(case))
            if block.default_node is not None:
                s.extend(cls.get_last_statements(block.default_node))
            else:
                s.append(None)
            return s
        if type(block) is IncompleteSwitchCaseNode:
            s = []
            for case in block.cases:
                s.extend(cls.get_last_statements(case))
            return s
        if type(block) is GraphRegion:
            # normally this should not happen. however, we have test cases that trigger this case.
            return []

        raise NotImplementedError

    #
    # Path predicate
    #

    EXC_COUNTER = 1000

    def _extract_predicate(self, src_block, dst_block, edge_type) -> claripy.ast.Bool:
        if edge_type == "exception":
            # TODO: THIS IS ABSOLUTELY A HACK. AT THIS MOMENT YOU SHOULD NOT ATTEMPT TO MAKE SENSE OF EXCEPTION EDGES.
            self.EXC_COUNTER += 1
            return self.claripy_ast_from_ail_condition(
                ailment.Expr.BinaryOp(
                    None,
                    "CmpEQ",
                    (
                        ailment.Expr.Register(0x400000 + self.EXC_COUNTER, None, self.EXC_COUNTER, 64),
                        ailment.Expr.Const(None, None, self.EXC_COUNTER, 64),
                    ),
                    False,
                ),
                ins_addr=dst_block.addr,
            )

        if type(src_block) is ConditionalBreakNode:
            # at this point ConditionalBreakNode stores a claripy AST
            bool_var = src_block.condition
            if src_block.target == dst_block.addr:
                return bool_var
            return claripy.Not(bool_var)

        if type(src_block) is GraphRegion:
            return claripy.true()

        # sometimes the last statement is the conditional jump. sometimes it's the first statement of the block
        if isinstance(src_block, ailment.Block) and src_block.statements and is_head_controlled_loop_block(src_block):
            last_stmt = next(
                iter(stmt for stmt in src_block.statements[:-1] if isinstance(stmt, ailment.Stmt.ConditionalJump)), None
            )
            assert last_stmt is not None
        else:
            last_stmt = self.get_last_statement(src_block)

        if last_stmt is None:
            return claripy.true()
        if type(last_stmt) is ailment.Stmt.Jump:
            if isinstance(last_stmt.target, ailment.Expr.Const):
                return claripy.true()
            # indirect jump
            target_ast = self.claripy_ast_from_ail_condition(last_stmt.target, ins_addr=last_stmt.tags["ins_addr"])
            return target_ast == dst_block.addr
        if type(last_stmt) is ailment.Stmt.ConditionalJump:
            bool_var = self.claripy_ast_from_ail_condition(
                last_stmt.condition, must_bool=True, ins_addr=last_stmt.tags["ins_addr"]
            )
            assert isinstance(bool_var, claripy.ast.Bool)
            if isinstance(last_stmt.true_target, ailment.Expr.Const) and last_stmt.true_target.value == dst_block.addr:
                return bool_var
            return claripy.Not(bool_var)

        return claripy.true()

    #
    # Expression conversion
    #

    def _convert_extract(self, hi, lo, expr, tags, memo=None):
        # ailment does not support Extract. We translate Extract to Convert and shift.
        if lo == 0:
            return ailment.Expr.Convert(
                None,
                expr.size(),
                hi + 1,
                False,
                self.convert_claripy_bool_ast(expr, memo=memo),
                **tags,
            )

        raise NotImplementedError("This case will be implemented once encountered.")

    def convert_claripy_bool_ast(self, cond, memo=None):
        """
        Convert recovered reaching conditions from claripy ASTs to ailment Expressions

        :return: None
        """

        if not isinstance(cond, claripy.ast.Base):
            return cond
        if memo is None:
            memo = {}
        if cond.hash() in memo:
            return memo[cond.hash()]
        r = self.convert_claripy_bool_ast_core(cond, memo)
        r = peephole_optimize_expr(r, self._peephole_expr_optimizations)
        memo[cond.hash()] = r
        return r

    def convert_claripy_bool_ast_core(self, cond, memo):
        if isinstance(cond, ailment.Expr.Expression):
            return cond

        if cond.op in {"BoolS", "BoolV"} and claripy.is_true(claripy.simplify(cond)):
            return ailment.Expr.Const(None, None, True, 1)
        if cond in self._condition_mapping:
            return self._condition_mapping[cond]
        if cond.op in {"BVS", "BoolS"} and cond.args[0] in self._condition_mapping:
            return self._condition_mapping[cond.args[0]]

        def _binary_op_reduce(op, args, tags, signed=False):
            r = None
            for arg in args:
                if r is None:
                    r = self.convert_claripy_bool_ast(arg, memo=memo)
                else:
                    r = ailment.Expr.BinaryOp(
                        None, op, (r, self.convert_claripy_bool_ast(arg, memo=memo)), signed, **tags
                    )
            return r

        def _unary_op_reduce(op, arg, tags):
            r = self.convert_claripy_bool_ast(arg, memo=memo)
            # TODO: Keep track of tags
            return ailment.Expr.UnaryOp(None, op, r, **tags)

        _mapping = {
            "Not": lambda cond_, tags: _unary_op_reduce("Not", cond_.args[0], tags),
            "__neg__": lambda cond_, tags: _unary_op_reduce("Not", cond_.args[0], tags),
            "__invert__": lambda cond_, tags: _unary_op_reduce("BitwiseNeg", cond_.args[0], tags),
            "And": lambda cond_, tags: _binary_op_reduce("LogicalAnd", cond_.args, tags),
            "Or": lambda cond_, tags: _binary_op_reduce("LogicalOr", cond_.args, tags),
            "__le__": lambda cond_, tags: _binary_op_reduce("CmpLE", cond_.args, tags, signed=True),
            "SLE": lambda cond_, tags: _binary_op_reduce("CmpLE", cond_.args, tags, signed=True),
            "__lt__": lambda cond_, tags: _binary_op_reduce("CmpLT", cond_.args, tags, signed=True),
            "SLT": lambda cond_, tags: _binary_op_reduce("CmpLT", cond_.args, tags, signed=True),
            "UGT": lambda cond_, tags: _binary_op_reduce("CmpGT", cond_.args, tags),
            "UGE": lambda cond_, tags: _binary_op_reduce("CmpGE", cond_.args, tags),
            "__gt__": lambda cond_, tags: _binary_op_reduce("CmpGT", cond_.args, tags, signed=True),
            "__ge__": lambda cond_, tags: _binary_op_reduce("CmpGE", cond_.args, tags, signed=True),
            "SGT": lambda cond_, tags: _binary_op_reduce("CmpGT", cond_.args, tags, signed=True),
            "SGE": lambda cond_, tags: _binary_op_reduce("CmpGE", cond_.args, tags, signed=True),
            "ULT": lambda cond_, tags: _binary_op_reduce("CmpLT", cond_.args, tags),
            "ULE": lambda cond_, tags: _binary_op_reduce("CmpLE", cond_.args, tags),
            "__eq__": lambda cond_, tags: _binary_op_reduce("CmpEQ", cond_.args, tags),
            "__ne__": lambda cond_, tags: _binary_op_reduce("CmpNE", cond_.args, tags),
            "__add__": lambda cond_, tags: _binary_op_reduce("Add", cond_.args, tags, signed=False),
            "__sub__": lambda cond_, tags: _binary_op_reduce("Sub", cond_.args, tags),
            "__mul__": lambda cond_, tags: _binary_op_reduce("Mul", cond_.args, tags),
            "__xor__": lambda cond_, tags: _binary_op_reduce("Xor", cond_.args, tags),
            "__or__": lambda cond_, tags: _binary_op_reduce("Or", cond_.args, tags, signed=False),
            "__and__": lambda cond_, tags: _binary_op_reduce("And", cond_.args, tags),
            "__lshift__": lambda cond_, tags: _binary_op_reduce("Shl", cond_.args, tags),
            "__rshift__": lambda cond_, tags: _binary_op_reduce("Sar", cond_.args, tags),
            "__floordiv__": lambda cond_, tags: _binary_op_reduce("Div", cond_.args, tags),
            "__mod__": lambda cond_, tags: _binary_op_reduce("Mod", cond_.args, tags),
            "LShR": lambda cond_, tags: _binary_op_reduce("Shr", cond_.args, tags),
            "BVV": lambda cond_, tags: ailment.Expr.Const(None, None, cond_.args[0], cond_.size(), **tags),
            "BoolV": lambda cond_, tags: (
                ailment.Expr.Const(None, None, True, 1, **tags)
                if cond_.args[0] is True
                else ailment.Expr.Const(None, None, False, 1, **tags)
            ),
            "Extract": lambda cond_, tags: self._convert_extract(*cond_.args, tags, memo=memo),
            "ZeroExt": lambda cond_, tags: _binary_op_reduce(
                "Concat", [claripy.BVV(0, cond_.args[0]), cond_.args[1]], tags
            ),
            "Concat": lambda cond_, tags: _binary_op_reduce("Concat", cond_.args, tags),
        }

        if cond.op in _mapping:
            if cond in self._ast2annotations:
                cond_tags = self._ast2annotations.get(cond)
            elif claripy.Not(cond) in self._ast2annotations:
                cond_tags = self._ast2annotations.get(claripy.Not(cond))
            else:
                cond_tags = {}
            return _mapping[cond.op](cond, cond_tags)
        raise NotImplementedError(
            f"Condition variable {cond} has an unsupported operator {cond.op}. Consider implementing."
        )

    def claripy_ast_from_ail_condition(
        self, condition, *, nobool: bool = False, must_bool: bool = False, ins_addr: int = 0
    ) -> claripy.ast.Bool | claripy.ast.Bits:
        # Unpack a condition all the way to the leaves
        if isinstance(condition, (claripy.ast.Bits, claripy.ast.Bool)):  # pylint:disable=isinstance-second-argument-not-valid-type
            return condition

        if isinstance(
            condition,
            (ailment.Expr.VEXCCallExpression, ailment.Expr.BasePointerOffset, ailment.Expr.ITE),
        ):
            return _dummy_bvs(condition, self._condition_mapping, must_bool=must_bool)
        if isinstance(condition, ailment.Expr.Call):
            return _dummy_bvs(
                condition,
                self._condition_mapping,
                name_suffix=hex(condition.tags.get("ins_addr", 0)),
                must_bool=must_bool,
            )
        if isinstance(condition, (ailment.Expr.Load, ailment.Expr.Register, ailment.Expr.VirtualVariable)):
            # does it have a variable associated?
            if condition.variable is not None:
                if condition.bits == 1:
                    var = claripy.BoolS(
                        f"ailexpr_{condition!r}-{condition.variable.ident}-{ins_addr:x}",
                        explicit_name=True,
                    )
                else:
                    var = claripy.BVS(
                        f"ailexpr_{condition!r}-{condition.variable.ident}-{ins_addr:x}",
                        condition.bits,
                        explicit_name=True,
                    )
            else:
                if condition.bits == 1:
                    var = claripy.BoolS(f"ailexpr_{condition!r}-{condition.idx}-{ins_addr:x}", explicit_name=True)
                else:
                    var = claripy.BVS(
                        f"ailexpr_{condition!r}-{condition.idx}-{ins_addr:x}", condition.bits, explicit_name=True
                    )
            self._condition_mapping[var.args[0]] = condition
            return var
        if isinstance(condition, ailment.expression.Extract):
            var_ = self.claripy_ast_from_ail_condition(condition.base, ins_addr=ins_addr)
            offset_expr = (
                self.claripy_ast_from_ail_condition(condition.offset, ins_addr=ins_addr).args[0]
                if not isinstance(condition.offset, ailment.expression.Const)
                else condition.offset.value
            )
            var = claripy.BVS(f"ailexpr_Extract({offset_expr}, {hash(var_)})", condition.bits, explicit_name=True)
            self._condition_mapping[var.args[0]] = condition
            return var
        if isinstance(condition, ailment.expression.Insert):
            var_ = self.claripy_ast_from_ail_condition(condition.base, ins_addr=ins_addr)
            value = self.claripy_ast_from_ail_condition(condition.value, ins_addr=ins_addr)
            offset_expr = (
                self.claripy_ast_from_ail_condition(condition.offset, ins_addr=ins_addr).args[0]
                if not isinstance(condition.offset, ailment.expression.Const)
                else condition.offset.value
            )
            var = claripy.BVS(
                f"ailexpr_Insert({offset_expr}, {hash(value)}, {hash(var_)})", condition.bits, explicit_name=True
            )
            self._condition_mapping[var.args[0]] = condition
            return var
        if isinstance(condition, ailment.Expr.Convert):
            # convert is special. if it generates a 1-bit variable, it should be treated as a BoolS
            if condition.to_bits == 1 and not nobool:
                var_ = self.claripy_ast_from_ail_condition(condition.operands[0], ins_addr=ins_addr)
                name = f"ailcond_Conv({condition.from_bits}->{condition.to_bits}, {hash(var_)})"
                var = claripy.BoolS(name, explicit_name=True)
            else:
                var_ = self.claripy_ast_from_ail_condition(condition.operands[0], ins_addr=ins_addr)
                name = f"ailexpr_Conv({condition.from_bits}->{condition.to_bits}, {hash(var_)})"
                var = claripy.BVS(name, condition.to_bits, explicit_name=True)
            self._condition_mapping[var.args[0]] = condition
            return var
        if isinstance(condition, ailment.Expr.Const):
            if condition.value is True or condition.value is False:
                var = claripy.BoolV(condition.value)
            else:
                var = claripy.BVV(condition.value, condition.bits)
                if condition.idx is not None:
                    # we do not want to lose track of this constant when it has idx
                    var = var.annotate(AILExprIdAnnotation())
                    self._condition_mapping[var] = condition
            if isinstance(var, claripy.ast.Bits) and var.size() == 1:
                var = claripy.true() if var.concrete_value == 1 else claripy.false()
            return var
        if isinstance(condition, ailment.Expr.Tmp):
            l.warning("Left-over ailment.Tmp variable %s.", condition)
            if condition.bits == 1:
                var = claripy.BoolS(f"ailtmp_{condition.tmp_idx}", explicit_name=True)
            else:
                var = claripy.BVS(f"ailtmp_{condition.tmp_idx}", condition.bits, explicit_name=True)
            self._condition_mapping[var.args[0]] = condition
            return var
        if isinstance(condition, ailment.Expr.MultiStatementExpression):
            # just cache it
            if condition.bits == 1:
                var = claripy.BoolS(f"mstmtexpr_{hash(condition)}", explicit_name=True)
            else:
                var = claripy.BVS(f"mstmtexpr_{hash(condition)}", condition.bits, explicit_name=True)
            self._condition_mapping[var.args[0]] = condition
            return var

        lambda_expr = _ail2claripy_op_mapping.get(condition.verbose_op)
        if lambda_expr is None:
            # fall back to op
            lambda_expr = _ail2claripy_op_mapping.get(condition.op)
        if lambda_expr is None:
            # fall back to the catch-all option
            l.debug(
                "Unsupported AIL expression operation %s (or verbose: %s). Fall back to the default catch-all dummy "
                "option. Consider implementing.",
                condition.op,
                condition.verbose_op,
            )
            lambda_expr = _ail2claripy_op_mapping["_DUMMY_"]
        r = lambda_expr(condition, self.claripy_ast_from_ail_condition, self._condition_mapping, ins_addr)

        if isinstance(r, claripy.ast.Bool) and nobool:
            r = claripy.BVS(f"ailexpr_from_bool_{r!r}", 1, explicit_name=True)
            self._condition_mapping[r.args[0]] = condition

        if r is NotImplemented:
            if condition.bits == 1 and not nobool:
                r = claripy.BoolS(f"ailexpr_{condition!r}", explicit_name=True)
            else:
                r = claripy.BVS(f"ailexpr_{condition!r}", condition.bits, explicit_name=True)
            self._condition_mapping[r.args[0]] = condition
        # don't lose tags
        self._ast2annotations[r] = condition.tags

        if isinstance(r, claripy.ast.BV) and r.size() == 1 and must_bool:
            # convert to a BoolS
            if r.op == "BVV":
                r = claripy.false() if r.args[0] == 0 else claripy.true()
            else:
                # r.op == "BVS"
                r = claripy.BoolS(f"bool_from_bv1_{r.args[0]}", explicit_name=True)
                self._condition_mapping[r.args[0]] = condition
        return r

    #
    # Expression simplification
    #

    @staticmethod
    def claripy_ast_to_sympy_expr(ast, memo=None):
        import sympy  # pylint:disable=import-outside-toplevel

        if ast.op == "And":
            return sympy.And(*(ConditionProcessor.claripy_ast_to_sympy_expr(arg, memo=memo) for arg in ast.args))
        if ast.op == "Or":
            return sympy.Or(*(ConditionProcessor.claripy_ast_to_sympy_expr(arg, memo=memo) for arg in ast.args))
        if ast.op == "Not":
            return sympy.Not(ConditionProcessor.claripy_ast_to_sympy_expr(ast.args[0], memo=memo))

        if ast.op in _UNIFIABLE_COMPARISONS:
            # unify comparisons to enable more simplification opportunities without going "deep" in sympy
            inverse_op = getattr(ast.args[0], _INVERSE_OPERATIONS[ast.op])
            return sympy.Not(ConditionProcessor.claripy_ast_to_sympy_expr(inverse_op(ast.args[1]), memo=memo))

        if memo is not None and ast in memo:
            return memo[ast]
        symbol = sympy.Symbol(str(hash(ast)))
        if memo is not None:
            memo[symbol] = ast
        return symbol

    @staticmethod
    def sympy_expr_to_claripy_ast(expr, memo: dict):
        import sympy  # pylint:disable=import-outside-toplevel

        if expr.is_Symbol:
            return memo[expr]
        if isinstance(expr, sympy.Or):
            return claripy.Or(*(ConditionProcessor.sympy_expr_to_claripy_ast(arg, memo) for arg in expr.args))
        if isinstance(expr, sympy.And):
            return claripy.And(*(ConditionProcessor.sympy_expr_to_claripy_ast(arg, memo) for arg in expr.args))
        if isinstance(expr, sympy.Not):
            return claripy.Not(ConditionProcessor.sympy_expr_to_claripy_ast(expr.args[0], memo))
        if isinstance(expr, sympy.logic.boolalg.BooleanTrue):
            return claripy.true()
        if isinstance(expr, sympy.logic.boolalg.BooleanFalse):
            return claripy.false()
        raise AngrRuntimeError("Unreachable reached")

    @staticmethod
    def simplify_condition(cond, depth_limit=8, variables_limit=8):
        import sympy  # pylint:disable=import-outside-toplevel

        memo = {}
        if cond.depth > depth_limit or len(cond.variables) > variables_limit:
            return cond
        sympy_expr = ConditionProcessor.claripy_ast_to_sympy_expr(cond, memo=memo)
        return ConditionProcessor.sympy_expr_to_claripy_ast(sympy.simplify_logic(sympy_expr, deep=False), memo)

    @staticmethod
    def _simplify_trivial_cases(cond):
        if cond.op == "And":
            new_args = []
            for arg in cond.args:
                claripy_simplified = claripy.simplify(arg)
                if claripy.is_true(claripy_simplified):
                    continue
                new_args.append(arg)
            return claripy.And(*new_args)

        return None

    @staticmethod
    def _revert_short_circuit_conditions(cond):
        # revert short-circuit conditions
        # !A||(A&&!B) ==> !(A&&B)

        if cond.op != "Or":
            return cond

        if len(cond.args) == 1:
            # redundant operator. get rid of it
            return cond.args[0]

        or_arg0, or_arg1 = cond.args[:2]
        if or_arg1.op == "And":
            pass
        elif or_arg0.op == "And":
            or_arg0, or_arg1 = or_arg1, or_arg0
        else:
            return cond

        not_a = or_arg0
        solver = claripy.SolverCacheless()

        if not_a.variables == or_arg1.args[0].variables:
            solver.add(not_a == or_arg1.args[0])
            not_b = or_arg1.args[1]
        elif not_a.variables == or_arg1.args[1].variables:
            solver.add(not_a == or_arg1.args[1])
            not_b = or_arg1.args[0]
        else:
            return cond

        if not solver.satisfiable():
            # found it!
            b = claripy.Not(not_b)
            a = claripy.Not(not_a)
            if len(cond.args) <= 2:
                return claripy.Not(claripy.And(a, b))
            return claripy.Or(claripy.Not(claripy.And(a, b)), *cond.args[2:])
        return cond

    @staticmethod
    def _fold_double_negations(cond):
        # !(!A) ==> A
        # !((!A) && (!B)) ==> A || B
        # !((!A) && B) ==> A || !B
        # !(A || B) ==> (!A && !B)

        if cond.op != "Not":
            return None
        if cond.args[0].op == "Not":
            return cond.args[0]

        if cond.args[0].op == "And" and len(cond.args[0].args) == 2:
            and_0, and_1 = cond.args[0].args
            if and_0.op == "Not" and and_1.op == "Not":
                return claripy.Or(and_0.args[0], and_1.args[0])

            if and_0.op == "Not":  # and_1.op != "Not"
                return claripy.Or(and_0.args[0], ConditionProcessor.simplify_condition(claripy.Not(and_1)))

        if cond.args[0].op == "Or" and len(cond.args[0].args) == 2:
            or_0, or_1 = cond.args[0].args
            return claripy.And(
                ConditionProcessor.simplify_condition(claripy.Not(or_0)),
                ConditionProcessor.simplify_condition(claripy.Not(or_1)),
            )

        return None

    @staticmethod
    def _extract_common_subexpressions(cond):
        def _expr_inside_collection(expr_, coll_) -> bool:
            return any(expr_ is ex_ for ex_ in coll_)

        # (A && B) || (A && C) => A && (B || C)
        if cond.op == "And":
            args = [ConditionProcessor._extract_common_subexpressions(arg) for arg in cond.args]
            if all(arg is None for arg in args):
                return None
            return claripy.And(*((arg if arg is not None else ori_arg) for arg, ori_arg in zip(args, cond.args)))

        if cond.op == "Or":
            args = [ConditionProcessor._extract_common_subexpressions(arg) for arg in cond.args]
            args = [(arg if arg is not None else ori_arg) for arg, ori_arg in zip(args, cond.args)]

            expr_ctrs = defaultdict(int)
            for arg in args:
                if arg.op == "And":
                    for subexpr in arg.args:
                        expr_ctrs[subexpr] += 1
                else:
                    expr_ctrs[arg] += 1

            common_exprs = []
            for expr, ctr in expr_ctrs.items():
                if ctr == len(args):
                    # found a common one
                    common_exprs.append(expr)

            if not common_exprs:
                return claripy.Or(*args)

            new_args = []
            for arg in args:
                if arg.op == "And":
                    new_subexprs = [
                        subexpr for subexpr in arg.args if not _expr_inside_collection(subexpr, common_exprs)
                    ]
                    new_args.append(claripy.And(*new_subexprs))
                elif arg in common_exprs:
                    continue
                else:
                    raise AngrRuntimeError("Unexpected behavior - you should never reach here")

            return claripy.And(*common_exprs, claripy.Or(*new_args))

        return None

    @staticmethod
    def _extract_terms(ast: claripy.ast.Bool) -> Generator[claripy.ast.Bool]:
        if ast.op == "And" or ast.op == "Or":
            for arg in ast.args:
                yield from ConditionProcessor._extract_terms(arg)
        elif ast.op == "Not":
            yield from ConditionProcessor._extract_terms(ast.args[0])
        else:
            yield ast

    @staticmethod
    def _replace_term_in_ast(
        ast: claripy.ast.Bool,
        r0: claripy.ast.Bool,
        r0_with: claripy.ast.Bool,
        r1: claripy.ast.Bool,
        r1_with: claripy.ast.Bool,
    ) -> claripy.ast.Bool:
        if ast.op == "And":
            return claripy.And(
                *(ConditionProcessor._replace_term_in_ast(arg, r0, r0_with, r1, r1_with) for arg in ast.args)
            )
        if ast.op == "Or":
            return claripy.Or(
                *(ConditionProcessor._replace_term_in_ast(arg, r0, r0_with, r1, r1_with) for arg in ast.args)
            )
        if ast.op == "Not":
            return claripy.Not(ConditionProcessor._replace_term_in_ast(ast.args[0], r0, r0_with, r1, r1_with))
        if ast is r0:
            return r0_with
        if ast is r1:
            return r1_with
        return ast

    @staticmethod
    def _remove_redundant_terms(cond):
        """
        Extract all terms and test for each term if its truism impacts the truism of the entire condition. If not, the
        term is redundant and can be replaced with a True.
        """

        all_terms = set()
        for term in ConditionProcessor._extract_terms(cond):
            if term not in all_terms:
                all_terms.add(term)

        negations = {}
        to_skip = set()
        all_terms_without_negs = set()
        for term in all_terms:
            if term in to_skip:
                continue
            neg = claripy.Not(term)
            if neg in all_terms:
                negations[term] = neg
                to_skip.add(neg)
                all_terms_without_negs.add(term)
            else:
                all_terms_without_negs.add(term)

        solver = claripy.SolverCacheless()
        for term in all_terms_without_negs:
            neg = negations.get(term)

            replaced_with_true = ConditionProcessor._replace_term_in_ast(
                cond, term, claripy.true(), neg, claripy.false()
            )
            sat0 = solver.satisfiable(
                extra_constraints=(
                    cond,
                    claripy.Not(replaced_with_true),
                )
            )
            sat1 = solver.satisfiable(
                extra_constraints=(
                    claripy.Not(cond),
                    replaced_with_true,
                )
            )
            if sat0 or sat1:
                continue

            replaced_with_false = ConditionProcessor._replace_term_in_ast(
                cond, term, claripy.false(), neg, claripy.true()
            )
            sat0 = solver.satisfiable(
                extra_constraints=(
                    cond,
                    claripy.Not(replaced_with_false),
                )
            )
            sat1 = solver.satisfiable(
                extra_constraints=(
                    claripy.Not(cond),
                    replaced_with_false,
                )
            )
            if sat0 or sat1:
                continue

            # TODO: Finish the implementation
            print(term, "is redundant")

    #
    # Graph processing
    #

    @staticmethod
    def _remove_crossing_edges_between_cases(
        graph: networkx.DiGraph, case_entry_to_switch_head: dict[int, int]
    ) -> networkx.DiGraph:
        starting_nodes = {node for node in graph if node.addr in case_entry_to_switch_head}
        if not starting_nodes:
            return graph

        traversed_nodes = set()
        edges_to_remove = set()
        for starting_node in starting_nodes:
            queue = [starting_node]
            while queue:
                src = queue.pop(0)
                traversed_nodes.add(src)
                successors = graph.successors(src)
                for succ in successors:
                    if succ in traversed_nodes:
                        # we should not traverse this node twice
                        if graph.out_degree(succ) > 0:
                            edges_to_remove.add((src, succ))
                        continue
                    if succ in starting_nodes:
                        # we do not want any jump from one node to a starting node
                        edges_to_remove.add((src, succ))
                        continue
                    traversed_nodes.add(src)
                    queue.append(succ)

        if not edges_to_remove:
            return graph

        # make a copy before modifying the graph
        graph = networkx.DiGraph(graph)
        graph.remove_edges_from(edges_to_remove)
        return graph

    #
    # Utils
    #

    def create_jump_target_var(self, jumptable_head_addr: int):
        return claripy.BVS(f"jump_table_{jumptable_head_addr:x}", self.arch.bits, explicit_name=True)



from __future__ import annotations

import os
import os.path
import struct
import tempfile

import angr
from angr.rustylib.fuzzer import (
    DeterministicMutator,
    Fuzzer,
    HavocMutator,
    InMemoryCorpus,
    OnDiskCorpus,
)
from tests.common import bin_location

# pylint: disable=missing-class-docstring,no-self-use

SHELLCODE = """
cmp al, 0x41
je path_a
cmp al, 0x42
je path_b
ret
path_a:
nop
ret
path_b:
nop
nop
ret
"""

# Shellcode with a crash path: 0x43 triggers an access to unmapped memory
SHELLCODE_WITH_CRASH = """
cmp al, 0x43
je crash_path
cmp al, 0x41
je path_a
ret
path_a:
nop
ret
crash_path:
mov rdi, 0xdeadbeef
mov byte ptr [rdi], 0
ret
"""

RETURN_ADDR = 0x100


def _apply_fn(state: angr.SimState, input: bytes):  # pylint: disable=redefined-builtin
    # Feed the fuzzed byte into rax so the binary can branch on it.
    state.regs.rax = input[0] if input else 0
    # Set the return address so the executor's breakpoint fires on normal return.
    # Must be inside the binary's mapped region but past the shellcode bytes.
    cc = state.project.factory.cc()
    cc.return_addr.set_value(state, RETURN_ADDR)


class TestFuzzer:
    def test_fuzzer(self):
        project = angr.load_shellcode(SHELLCODE, "amd64")
        base_state = project.factory.entry_state()
        corpus = InMemoryCorpus.from_list([b"\x00", b"A", b"B", b"C"])
        solutions = InMemoryCorpus()

        fuzzer = Fuzzer(base_state, corpus, solutions, _apply_fn, 0, 0, max_mutations=2)

        new_corpus_entry = fuzzer.run_once()
        live_corpus = fuzzer.corpus()
        assert isinstance(live_corpus, InMemoryCorpus)
        assert 0 <= new_corpus_entry < len(live_corpus)

    def test_fuzzer_ondisk(self):
        project = angr.load_shellcode(SHELLCODE, "amd64")
        base_state = project.factory.entry_state()

        with tempfile.TemporaryDirectory() as tmp_path:
            corpus_dir = os.path.join(tmp_path, "corpus")
            os.makedirs(corpus_dir)
            corpus = OnDiskCorpus(corpus_dir)
            for seed in [b"\x00", b"A", b"B", b"C"]:
                corpus.add(seed)
            solutions_dir = os.path.join(tmp_path, "solutions")
            os.makedirs(solutions_dir)
            solutions = OnDiskCorpus(solutions_dir)

            fuzzer = Fuzzer(base_state, corpus, solutions, _apply_fn, 0, 0, max_mutations=2)

            new_corpus_entry = fuzzer.run_once()
            live_corpus = fuzzer.corpus()
            assert isinstance(live_corpus, OnDiskCorpus)
            assert 0 <= new_corpus_entry < len(live_corpus)
            live_solutions = fuzzer.solutions()
            assert isinstance(live_solutions, OnDiskCorpus)

    def test_fuzzer_mixed_inmem_corpus_ondisk_solutions(self):
        project = angr.load_shellcode(SHELLCODE, "amd64")
        base_state = project.factory.entry_state()

        with tempfile.TemporaryDirectory() as tmp_path:
            corpus = InMemoryCorpus.from_list([b"\x00", b"A", b"B", b"C"])
            solutions_dir = os.path.join(tmp_path, "solutions_mixed1")
            os.makedirs(solutions_dir)
            solutions = OnDiskCorpus(solutions_dir)

            fuzzer = Fuzzer(base_state, corpus, solutions, _apply_fn, 0, 0, max_mutations=2)
            new_corpus_entry = fuzzer.run_once()
            live_corpus = fuzzer.corpus()
            assert isinstance(live_corpus, InMemoryCorpus)
            assert 0 <= new_corpus_entry < len(live_corpus)
            live_solutions = fuzzer.solutions()
            assert isinstance(live_solutions, OnDiskCorpus)

    def test_fuzzer_mixed_ondisk_corpus_inmem_solutions(self):
        project = angr.load_shellcode(SHELLCODE, "amd64")
        base_state = project.factory.entry_state()

        with tempfile.TemporaryDirectory() as tmp_path:
            corpus_dir = os.path.join(tmp_path, "corpus_mixed2")
            os.makedirs(corpus_dir)
            corpus = OnDiskCorpus(corpus_dir)
            for seed in [b"\x00", b"A", b"B", b"C"]:
                corpus.add(seed)
            solutions = InMemoryCorpus()

            fuzzer = Fuzzer(base_state, corpus, solutions, _apply_fn, 0, 0, max_mutations=2)
            new_corpus_entry = fuzzer.run_once()
            live_corpus = fuzzer.corpus()
            assert isinstance(live_corpus, OnDiskCorpus)
            assert 0 <= new_corpus_entry < len(live_corpus)
            live_solutions = fuzzer.solutions()
            assert isinstance(live_solutions, InMemoryCorpus)

    def test_havoc_mutator_config(self):
        """Test that HavocMutator with configuration options works."""
        project = angr.load_shellcode(SHELLCODE, "amd64")
        base_state = project.factory.entry_state()
        corpus = InMemoryCorpus.from_list([b"\x00", b"A", b"B", b"C"])
        solutions = InMemoryCorpus()

        mutator = HavocMutator(max_stack_pow=2)
        fuzzer = Fuzzer(
            base_state,
            corpus,
            solutions,
            _apply_fn,
            0,
            0,
            max_mutations=2,
            mutator=mutator,
        )

        new_corpus_entry = fuzzer.run_once()
        live_corpus = fuzzer.corpus()
        assert isinstance(live_corpus, InMemoryCorpus)
        assert 0 <= new_corpus_entry < len(live_corpus)

    def test_deterministic_mutator_finds_solution(self):
        """Test that a deterministic mutator producing a crash value yields a solution."""
        project = angr.load_shellcode(SHELLCODE_WITH_CRASH, "amd64")
        base_state = project.factory.entry_state()

        corpus = InMemoryCorpus.from_list([b"\x00"])
        solutions = InMemoryCorpus()

        # The deterministic mutator will produce 0x43 ('C') which triggers the crash path
        mutator = DeterministicMutator([b"\x43"])
        fuzzer = Fuzzer(
            base_state,
            corpus,
            solutions,
            _apply_fn,
            0,
            0,
            max_mutations=1,
            mutator=mutator,
        )

        fuzzer.run_once()
        live_solutions = fuzzer.solutions()
        assert len(live_solutions) >= 1, "Expected at least one solution from crash path"

    def test_deterministic_mutator_no_crash(self):
        """Test that a deterministic mutator with non-crash values does not produce solutions."""
        project = angr.load_shellcode(SHELLCODE_WITH_CRASH, "amd64")
        base_state = project.factory.entry_state()

        corpus = InMemoryCorpus.from_list([b"\x00"])
        solutions = InMemoryCorpus()

        # Only produce values that do NOT trigger the crash path
        mutator = DeterministicMutator([b"\x41"])
        fuzzer = Fuzzer(
            base_state,
            corpus,
            solutions,
            _apply_fn,
            0,
            0,
            max_mutations=1,
            mutator=mutator,
        )

        fuzzer.run_once()
        live_solutions = fuzzer.solutions()
        assert len(live_solutions) == 0, "path_a should not crash"

    def test_deterministic_mutator_sequence(self):
        """Test that a deterministic mutator cycles through its sequence and finds solutions."""
        project = angr.load_shellcode(SHELLCODE_WITH_CRASH, "amd64")
        base_state = project.factory.entry_state()

        corpus = InMemoryCorpus.from_list([b"\x00"])
        solutions = InMemoryCorpus()

        # Cycle: first produces 0x41 (path_a, no crash), then 0x43 (crash_path)
        mutator = DeterministicMutator([b"\x41", b"\x43"])
        fuzzer = Fuzzer(
            base_state,
            corpus,
            solutions,
            _apply_fn,
            0,
            0,
            max_mutations=1,
            mutator=mutator,
        )

        # First iteration: mutator produces 0x41 -> path_a -> no crash
        fuzzer.run_once()
        solutions_after_first = fuzzer.solutions()
        assert len(solutions_after_first) == 0, "path_a should not crash"

        # Second iteration: mutator produces 0x43 -> crash_path -> solution
        fuzzer.run_once()
        solutions_after_second = fuzzer.solutions()
        assert len(solutions_after_second) >= 1, "crash_path should produce a solution"

    def test_vuln_stacksmash_deterministic(self):
        """Test that DeterministicMutator detects a stack buffer overflow in a real binary.

        vuln_stacksmash has main():
            sub rsp, 0x70        // 112-byte buffer
            read(0, buf, 0x800)  // reads up to 2048 bytes into 112-byte buffer
            leave; ret

        We start execution after the read() call and write fuzzed bytes directly
        into the buffer.  Inputs longer than 120 bytes overwrite the return
        address, causing MEMORY_ERROR on ret → CrashFeedback → solution.
        """
        bin_path = os.path.join(bin_location, "tests", "x86_64", "vuln_stacksmash")
        project = angr.Project(bin_path, auto_load_libs=False)

        AFTER_READ = 0x400505  # mov eax, 0 (instruction after call read@plt)
        STACK_RET = 0x400100  # breakpoint address, within binary's mapped region

        # Allocate and fully materialise a stack page so icicle can sync it.
        # (concrete_load returns empty for sparse pages, causing lost writes.)
        STACK_PAGE = 0x651000
        base_state = project.factory.blank_state(
            addr=AFTER_READ,
            add_options={
                angr.sim_options.ZERO_FILL_UNCONSTRAINED_MEMORY,
                angr.sim_options.ZERO_FILL_UNCONSTRAINED_REGISTERS,
            },
        )
        base_state.memory.map_region(STACK_PAGE, 0x1000, 7)  # rwx
        base_state.memory.store(STACK_PAGE, b"\x00" * 0x1000)  # materialise

        # Set up the stack frame as if main's prologue already ran:
        #   push rbp; mov rbp, rsp; sub rsp, 0x70
        RBP = STACK_PAGE + 0xFF0  # near top of page
        RSP = RBP - 0x70

        base_state.regs.rbp = RBP
        base_state.regs.rsp = RSP

        def stacksmash_apply_fn(state: angr.SimState, input_bytes: bytes):
            # Reset the saved rbp and return address each iteration
            state.memory.store(RBP, struct.pack("<Q", 0))  # saved rbp
            state.memory.store(RBP + 8, struct.pack("<Q", STACK_RET))  # return addr
            # Write fuzzed input into the buffer at [rbp - 0x70].
            # >120 bytes overflows past saved rbp (8) into the return address.
            state.memory.store(RBP - 0x70, input_bytes)
            # Tell the executor where to set its breakpoint
            cc = state.project.factory.cc()
            cc.return_addr.set_value(state, STACK_RET)

        corpus = InMemoryCorpus.from_list([b"\x00"])
        solutions = InMemoryCorpus()

        # 128 bytes of 'A': overwrites buffer (112) + saved rbp (8) + return addr (8)
        crash_input = b"A" * 128
        mutator = DeterministicMutator([crash_input])
        fuzzer = Fuzzer(
            base_state,
            corpus,
            solutions,
            stacksmash_apply_fn,
            0,
            0,
            max_mutations=1,
            mutator=mutator,
        )

        fuzzer.run_once()
        live_solutions = fuzzer.solutions()
        assert len(live_solutions) >= 1, "Stack buffer overflow should produce a solution"



# pylint:disable=missing-class-docstring
from __future__ import annotations
import angr

from angr.sim_type import SimStruct, SimTypeLong, SimTypeFixedSizeArray, SimTypeShort, SimTypeInt

# struct sysinfo {
#    long uptime;             /* Seconds since boot */
#    unsigned long loads[3];  /* 1, 5, and 15 minute load averages */
#    unsigned long totalram;  /* Total usable main memory size */
#    unsigned long freeram;   /* Available memory size */
#    unsigned long sharedram; /* Amount of shared memory */
#    unsigned long bufferram; /* Memory used by buffers */
#    unsigned long totalswap; /* Total swap space size */
#    unsigned long freeswap;  /* Swap space still available */
#    unsigned short procs;    /* Number of current processes */
#    unsigned long totalhigh; /* Total high memory size */
#    unsigned long freehigh;  /* Available high memory size */
#    unsigned int mem_unit;   /* Memory unit size in bytes */
# }
sysinfo_ty = SimStruct(
    {
        "uptime": SimTypeLong(signed=True),
        "loads": SimTypeFixedSizeArray(SimTypeLong(signed=False), 3),
        "totalram": SimTypeLong(signed=False),
        "freeram": SimTypeLong(signed=False),
        "sharedram": SimTypeLong(signed=False),
        "bufferram": SimTypeLong(signed=False),
        "totalswap": SimTypeLong(signed=False),
        "freeswap": SimTypeLong(signed=False),
        "procs": SimTypeShort(signed=False),
        "totalhigh": SimTypeLong(signed=False),
        "freehigh": SimTypeLong(signed=False),
        "mem_unit": SimTypeInt(signed=False),
    },
    name="sysinfo",
    pack=False,
    align=None,
)


class sysinfo(angr.SimProcedure):
    def run(self, info):  # pylint: disable=arguments-differ
        value = {
            "uptime": 1234567,
            "loads": [20100, 22000, 15000],
            "totalram": 1024**2,
            "freeram": 1024**2 // 4,
            "sharedram": 1024**2 // 4,
            "bufferram": 1024**2 // 4,
            "totalswap": 1024**2,
            "freeswap": 1024**2 // 2,
            "procs": 533,
            "totalhigh": 11,
            "freehigh": 12,
            "mem_unit": 13,
        }
        sysinfo_ty.with_arch(self.arch).store(self.state, info, value)
        return 0



from __future__ import annotations
from typing import TYPE_CHECKING, Generic
import logging


from angr.engines.light.engine import BlockType, DataType_co, StateType

from angr.engines.light import SimEngineLight
from angr.errors import SimEngineError
from angr.project import Project
from angr.misc.testing import is_testing

if TYPE_CHECKING:
    from angr.analyses.reaching_definitions.reaching_definitions import ReachingDefinitionsModel

l = logging.getLogger(name=__name__)


class SimEnginePropagatorBaseMixin(
    Generic[StateType, DataType_co, BlockType], SimEngineLight[StateType, DataType_co, BlockType, StateType]
):  # pylint:disable=abstract-method
    """
    The base class for the propagator VEX engine.
    """

    def __init__(
        self,
        project: Project,
        stack_pointer_tracker=None,
        propagate_tmps=True,
        reaching_definitions: ReachingDefinitionsModel | None = None,
        bp_as_gpr: bool = False,
    ):
        super().__init__(project)

        # Used in the VEX engine
        self.arch = project.arch
        self.base_state = None
        self._load_callback = None
        self._propagate_tmps: bool = propagate_tmps
        self._reaching_definitions = reaching_definitions
        self.bp_as_gpr = bp_as_gpr

        # Used in the AIL engine
        self._stack_pointer_tracker = stack_pointer_tracker

        self._multi_occurrence_registers = None

    def process(
        self, state: StateType, *, block: BlockType | None = None, base_state=None, load_callback=None, **kwargs
    ) -> StateType:
        self.base_state = base_state
        self._load_callback = load_callback
        try:
            result_state = super().process(state, block=block, **kwargs)
        except SimEngineError as ex:
            if kwargs.pop("fail_fast", is_testing) is True:
                raise ex
            l.error(ex, exc_info=True)
            result_state = state

        return result_state



from __future__ import annotations
from typing import Any
import logging

import networkx

import angr.ailment as ailment
from angr.ailment import Block, AILBlockRewriter
from angr.ailment.statement import Jump, ConditionalJump, Assignment, Return, Label, SideEffectStatement
from angr.ailment.expression import Const, Phi, VirtualVariable

from angr.utils.ail import is_phi_assignment
from angr.analyses.decompiler.condition_processor import ConditionProcessor, EmptyBlockNotice
from angr.analyses.decompiler.graph_region import GraphRegion
from angr.analyses.decompiler.utils import remove_labels, to_ail_supergraph, calls_in_graph
from angr.analyses.decompiler.structuring.structurer_nodes import MultiNode, ConditionNode
from angr.analyses.decompiler.region_identifier import RegionIdentifier

_l = logging.getLogger(name=__name__)


class FreshVirtualVariableRewriter(AILBlockRewriter):
    """
    Helper class to rewrite virtual variables so that they will use fresh virtual variables.
    """

    def __init__(self, vvar_id_start: int, vvar_mapping: dict[int, int]):
        super().__init__()
        self.vvar_idx = vvar_id_start
        self.vvar_mapping = vvar_mapping
        self.new_block = None

    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Block | None):
        new_stmt = super()._handle_Assignment(stmt_idx, stmt, block)
        dst = new_stmt.dst if new_stmt is not None else stmt.dst
        src = new_stmt.src if new_stmt is not None else stmt.src
        if isinstance(dst, VirtualVariable):
            self.vvar_mapping[dst.varid] = self.vvar_idx
            self.vvar_idx += 1

            dst = VirtualVariable(
                dst.idx,
                self.vvar_mapping[dst.varid],
                dst.bits,
                dst.category,
                dst.oident,
                variable=dst.variable,
                variable_offset=dst.variable_offset,
                **dst.tags,
            )

            return Assignment(stmt.idx, dst, src, **stmt.tags)

        return new_stmt

    def _handle_VirtualVariable(
        self, expr_idx: int, expr: VirtualVariable, stmt_idx: int, stmt, block: Block | None
    ) -> VirtualVariable:
        if expr.varid in self.vvar_mapping:
            return VirtualVariable(
                expr.idx,
                self.vvar_mapping[expr.varid],
                expr.bits,
                expr.category,
                expr.oident,
                variable=expr.variable,
                variable_offset=expr.variable_offset,
                **expr.tags,
            )
        return expr


class ReturnDuplicatorBase:
    """
    The base class for implementing Return Duplication as described in the SAILR paper.
    This base class describes the general algorithm for duplicating return regions in a graph.
    """

    # pylint:disable=unused-argument
    def __init__(
        self,
        func,
        manager,
        *,
        vvar_id_start: int,
        max_calls_in_regions: int = 2,
        minimize_copies_for_regions: bool = True,
        ri: RegionIdentifier | None = None,
        scratch: dict[str, Any] | None = None,
        max_func_blocks: int = 1500,
    ):
        self._max_calls_in_region = max_calls_in_regions
        self._minimize_copies_for_regions = minimize_copies_for_regions
        self._supergraph = None

        # this should also be set by the optimization passes initer
        self.scratch = scratch if scratch is not None else {}
        self._func = func
        self._manager = manager
        self._ri: RegionIdentifier | None = ri
        self.vvar_id_start = vvar_id_start
        self._max_func_blocks = max_func_blocks

    def next_node_idx(self) -> int:
        node_idx = self.scratch.get("returndup_node_idx", 0) + 1
        self.scratch["returndup_node_idx"] = node_idx
        return node_idx

    #
    # must implement these methods
    #

    def _should_duplicate_dst(self, src, dst, graph, dst_is_const_ret=False) -> bool:
        raise NotImplementedError

    #
    # main analysis
    #

    def _check(self):
        # is this function too large?
        if len(self._func.block_addrs_set) > self._max_func_blocks:
            return False, None
        # does this function have end points?
        return bool(self._func.endpoints), None

    def _analyze_core(self, graph: networkx.DiGraph) -> bool:
        """
        This function does the core checks and duplications to the graph passed.
        The return value is True if the graph was changed.
        """
        graph_changed = False
        endnode_regions = self._find_endnode_regions(graph)

        if self._minimize_copies_for_regions:
            # perform a second pass to minimize the number of copies by doing only a single copy
            # for connected in_edges that form a region
            endnode_regions = self._copy_connected_edge_components(endnode_regions, graph)

        # refresh the supergraph
        self._supergraph = to_ail_supergraph(graph)
        for region_head, (in_edges, region) in endnode_regions.items():
            is_single_const_ret_region = self._is_simple_return_graph(region)
            dup_pred_nodes = []
            # duplicate the entire region if at least (N-2) in-edges for the region head is deemed should be duplicated.
            # otherwise we only duplicate the edges that should be duplicated
            for in_edge in in_edges:
                pred_node = in_edge[0]
                if self._should_duplicate_dst(
                    pred_node, region_head, graph, dst_is_const_ret=is_single_const_ret_region
                ):
                    dup_pred_nodes.append(pred_node)

            dup_count = len(dup_pred_nodes)
            dup_all = dup_count >= len(in_edges) - 2 > 0
            if dup_all:
                for pred_node in sorted((in_edge[0] for in_edge in in_edges), key=lambda x: x.addr):
                    # every eligible pred gets a new region copy
                    self._copy_region([pred_node], region_head, region, graph)
                    graph_changed = True
            else:
                for pred_node in dup_pred_nodes:
                    self._copy_region([pred_node], region_head, region, graph)
                    graph_changed = True

            if region_head in graph and graph.in_degree(region_head) == 0:
                graph.remove_nodes_from(region)
                graph_changed = True

        return graph_changed

    def _find_endnode_regions(self, graph) -> dict[Any, tuple[list[tuple[Any, Any]], networkx.DiGraph]]:
        """
        Find all the regions that contain a node with no successors. These are the "end nodes" of the graph.
        """
        endnodes = [node for node in graph.nodes() if graph.out_degree[node] == 0]

        # to_update is keyed by the region head.
        # this is because different end nodes may lead to the same region head: consider the case of the typical "fork"
        # region where stack canary is checked in x86-64 binaries.
        end_node_regions: dict[Any, tuple[list[tuple[Any, Any]], networkx.DiGraph]] = {}

        for end_node in endnodes:
            in_edges = list(graph.in_edges(end_node))

            if len(in_edges) > 1:
                region = networkx.DiGraph()
                region.add_node(end_node)
                region_head = end_node
            elif len(in_edges) == 1:
                # back-trace until it reaches a node with two predecessors
                region, region_head = self._single_entry_region(graph, end_node)
                tmp_in_edges = graph.in_edges(region_head)
                # remove in_edges that are coming from a node inside the region
                in_edges = []
                for src, dst in tmp_in_edges:
                    if src not in region:
                        in_edges.append((src, dst))
            else:  # len(in_edges) == 0
                continue

            # region and in_edge might have been updated. re-check
            if not in_edges:
                # this is a single connected component in the graph
                # no need to duplicate anything
                continue
            if len(in_edges) == 1:
                # there is no need to duplicate it
                continue

            if any(self._is_indirect_jump_ailblock(src) for src, _ in in_edges):
                continue

            # to assure we are not copying like crazy, set a max amount of code (which is estimated in calls)
            # that can be copied in a region
            if calls_in_graph(region) > self._max_calls_in_region:
                continue

            end_node_regions[region_head] = in_edges, region

        return end_node_regions

    def _copy_region(self, pred_nodes: list[Block], region_head, region, graph):
        # copy the entire return region
        copies: dict[Block, Block] = {}
        queue: list[tuple[Block, Block]] = [(pred_node, region_head) for pred_node in pred_nodes]
        vvar_mapping: dict[int, int] = {}
        while queue:
            pred, node = queue.pop(0)
            if node in copies:
                node_copy = copies[node]
            else:
                node_copy = node.deep_copy(self._manager)
                if node is region_head:
                    node_copy = self._copy_node_and_update_phi_variables(node_copy, pred)
                node_copy = self._use_fresh_virtual_variables(node_copy, vvar_mapping)
                node_copy.idx = self.next_node_idx()
                self._fix_copied_node_labels(node_copy)
                copies[node] = node_copy

            # modify Jump.target_idx and ConditionalJump.{true,false}_target_idx accordingly
            graph.add_edge(pred, node_copy)
            try:
                last_stmt = ConditionProcessor.get_last_statement(pred)
                if isinstance(last_stmt, Jump):
                    if isinstance(last_stmt.target, Const) and last_stmt.target.value == node_copy.addr:
                        updated_last_stmt = Jump(
                            last_stmt.idx, last_stmt.target, target_idx=node_copy.idx, **last_stmt.tags
                        )
                        pred.statements[-1] = updated_last_stmt
                elif isinstance(last_stmt, ConditionalJump):
                    if isinstance(last_stmt.true_target, Const) and last_stmt.true_target.value == node_copy.addr:
                        updated_last_stmt = ConditionalJump(
                            last_stmt.idx,
                            last_stmt.condition,
                            last_stmt.true_target,
                            last_stmt.false_target,
                            true_target_idx=node_copy.idx,
                            false_target_idx=last_stmt.false_target_idx,
                            **last_stmt.tags,
                        )
                        pred.statements[-1] = updated_last_stmt
                    elif isinstance(last_stmt.false_target, Const) and last_stmt.false_target.value == node_copy.addr:
                        updated_last_stmt = ConditionalJump(
                            last_stmt.idx,
                            last_stmt.condition,
                            last_stmt.true_target,
                            last_stmt.false_target,
                            true_target_idx=last_stmt.true_target_idx,
                            false_target_idx=node_copy.idx,
                            **last_stmt.tags,
                        )
                        pred.statements[-1] = updated_last_stmt
            except EmptyBlockNotice:
                pass

            for succ in region.successors(node):
                queue.append((node_copy, succ))

        for pred_node in pred_nodes:
            # delete the old edge to the return node
            graph.remove_edge(pred_node, region_head)
            # update phi variables at the beginning of region_head
            self._update_phi_variables_after_removing_predecessor(region_head, pred_node)

    @staticmethod
    def _copy_node_and_update_phi_variables(node: Block, pred: Block) -> Block:
        stmts = []
        for stmt in node.statements:
            if isinstance(stmt, Assignment) and isinstance(stmt.src, Phi) and isinstance(stmt.dst, VirtualVariable):
                # pick the variable from the correct source
                vvar_src = next(iter(vvar for src, vvar in stmt.src.src_and_vvars if src == (pred.addr, pred.idx)), ...)
                if vvar_src is ...:
                    _l.warning(
                        "Cannot find the virtual variable in Phi expression %r that corresponds with node %#x-%s.",
                        stmt.src,
                        pred.addr,
                        pred.idx,
                    )
                elif vvar_src is None:
                    # not used in this branch. drop this statement
                    continue
                else:
                    new_stmt = Assignment(stmt.idx, stmt.dst, vvar_src, **stmt.tags)
                    stmts.append(new_stmt)
                    continue
            stmts.append(stmt)

        return Block(node.addr, node.original_size, statements=stmts, idx=node.idx)

    @staticmethod
    def _update_phi_variables_after_removing_predecessor(node: Block, pred: Block) -> None:
        for idx in range(len(node.statements)):  # pylint:disable=consider-using-enumerate
            stmt = node.statements[idx]
            if isinstance(stmt, Assignment) and isinstance(stmt.src, Phi) and isinstance(stmt.dst, VirtualVariable):
                # remove the variable from the specified source
                new_src_and_vvars = [
                    (src, vvar) for src, vvar in stmt.src.src_and_vvars if src != (pred.addr, pred.idx)
                ]
                new_phi = Phi(stmt.src.idx, stmt.src.bits, new_src_and_vvars, **stmt.src.tags)
                node.statements[idx] = Assignment(stmt.idx, stmt.dst, new_phi, **stmt.tags)

    def _use_fresh_virtual_variables(self, node: Block, vvar_map: dict[int, int]) -> Block:
        rewriter = FreshVirtualVariableRewriter(self.vvar_id_start, vvar_map)
        rewriter.walk(node)
        self.vvar_id_start = rewriter.vvar_idx + 1
        return rewriter.new_block if rewriter.new_block is not None else node

    def _copy_connected_edge_components(
        self, endnode_regions: dict[Any, tuple[list[tuple[Any, Any]], networkx.DiGraph]], graph: networkx.DiGraph
    ):
        updated_regions = endnode_regions.copy()
        assert self._ri is not None
        assert isinstance(self._ri.region, GraphRegion)
        all_region_block_addrs = list(self._find_block_sets_in_all_regions(self._ri.region).values())
        for region_head, (in_edges, region) in endnode_regions.items():
            is_single_const_ret_region = self._is_simple_return_graph(region)
            pred_nodes = [src for src, _ in in_edges]
            pred_subgraph = networkx.subgraph(graph, pred_nodes)
            components = list(networkx.weakly_connected_components(pred_subgraph))
            multi_node_components = [c for c in components if len(c) > 1]
            if not multi_node_components:
                continue

            # find components that have a node that should be duplicated
            candidate_components = []
            for nodes in multi_node_components:
                if any(
                    self._should_duplicate_dst(n, region_head, graph, dst_is_const_ret=is_single_const_ret_region)
                    for n in nodes
                ):
                    candidate_components.append(nodes)
            if not candidate_components:
                continue

            # we can only handle instances where components do not overlap
            overlapping_comps = set()
            for component in candidate_components:
                overlapping_comps &= component
            if overlapping_comps:
                continue

            # every component needs to form its own region with ONLY those nodes in the region
            duplicatable_components = []
            for component in candidate_components:
                comp_addrs = {n.addr for n in component}
                if comp_addrs in all_region_block_addrs:
                    duplicatable_components.append(component)

            new_in_edges = in_edges
            for nodes in duplicatable_components:
                self._copy_region(nodes, region_head, region, graph)
                if region_head in graph and graph.in_degree(region_head) == 0:
                    graph.remove_nodes_from(region)

                # update the in_edges to remove any nodes that have been copied
                new_in_edges = [edge for edge in new_in_edges if edge[0] not in nodes]

            if not new_in_edges:
                del updated_regions[region_head]
            else:
                updated_regions[region_head] = new_in_edges, region

        return updated_regions

    @staticmethod
    def _is_simple_return_graph(graph: networkx.DiGraph, max_assigns=1):
        """
        Checks if the provided graph is a graph that ONLY contains a "simple" return.
        If there were absolutely no bugs in angr, we could just check that a single return block exists.
        However, due to some propagation bugs, these cases can all happen and are all valid:
        1. [Jmp] -> [Jmp] -> [Ret]
        2. [Jmp] -> [Jmp, x=0] -> [Ret x]
        3. [Jmp] -> [Jmp, x=rdi] -> [Ret x]

        To deal with this, we need to do the sketchy checks we do below.
        """
        labeless_graph = to_ail_supergraph(remove_labels(graph))
        nodes = list(labeless_graph.nodes())
        if not nodes:
            return False

        # check if the graph is a single successor chain
        if not all(labeless_graph.out_degree[n] <= 1 for n in nodes):
            return False

        # collect the statements from the top node, make sure one exists
        root_nodes = [n for n in nodes if labeless_graph.in_degree(n) == 0]
        if len(root_nodes) != 1:
            return False

        root_node = root_nodes[0]
        queue = [root_node]
        stmts = []
        while queue:
            node = queue.pop(0)
            succs = list(labeless_graph.successors(node))
            queue += succs
            if node.statements:
                stmts += node.statements

        # all statements must be either a return, a jump, an assignment, or a side-effect statement (e.g. a call)
        type_white_list = (Return, Jump, Assignment, SideEffectStatement)
        for stmt in stmts:
            if not isinstance(stmt, type_white_list):
                return False

        # gather all assignments
        assignments = [s for s in stmts if isinstance(s, Assignment) and not is_phi_assignment(s)]
        has_assign = len(assignments) > 0
        if len(assignments) > max_assigns:
            return False

        # gather return stmts
        ret_stmt = stmts[-1]
        if not isinstance(ret_stmt, Return):
            if isinstance(ret_stmt, Jump):
                _l.warning("Found a jump at the end of a return graph, did function analysis fail?")
            return False

        ret_exprs = ret_stmt.ret_exprs
        # must be 1 or none
        if ret_exprs and len(ret_exprs) > 1:
            return False

        if not ret_exprs:
            # a simple return statement that does not carry any value or variable to return
            return True

        ret_expr = ReturnDuplicatorBase.unwrap_conv(ret_exprs[0])
        # check if ret_expr is a virtual variable or not
        if not isinstance(ret_expr, (VirtualVariable, Const)):
            return False

        # stop early if there are no assignments at all and just jumps and rets, or a const ret
        if not has_assign:
            return True

        assign: Assignment = assignments[0]
        # const assignments are valid
        if isinstance(assign.src, Const):
            valid_assignment = ret_expr and ret_expr.likes(assign.dst)
        # assignments to registers from the stack are valid, since cases of these assignments
        # pop up across optimized binaries
        elif (
            isinstance(assign.dst, ailment.expression.VirtualVariable)
            and assign.dst.was_reg
            and isinstance(assign.src, ailment.expression.VirtualVariable)
            and assign.src.was_stack
            and isinstance(assign.src.stack_offset, int)
        ):
            valid_assignment = True
        else:
            valid_assignment = False

        return valid_assignment

    @staticmethod
    def _single_entry_region(graph, end_node) -> tuple[networkx.DiGraph, Any]:
        """
        Back track on the graph from `end_node` and find the longest chain of nodes where each node has only one
        predecessor and one successor (the second-to-last node may have two successors to account for the typical
        stack-canary-detection logic).

        :param end_node:    A node in the graph.
        :return:            A graph of nodes where the first node either has no predecessors or at least two
                            predecessors.
        """

        def _is_fork_node(node_) -> bool:
            """
            Check if the node and its successors form a "fork" region. A "fork" region is a region where:
            - The entry node has two successors,
            - Each successor has only the entry node as its predecessor.
            - Each successor has no successors.
            """

            succs = list(graph.successors(node_))
            if len(succs) != 2:
                return False
            for succ in succs:
                if graph.in_degree[succ] != 1:
                    return False
                if graph.out_degree[succ] != 0:
                    return False
            return True

        region = networkx.DiGraph()
        region.add_node(end_node)

        traversed = {end_node}
        region_head = end_node
        while True:
            preds = list(graph.predecessors(region_head))
            if len(preds) != 1:
                break
            second_to_last_node = region_head is end_node

            pred_node = preds[0]

            if pred_node in traversed:
                break

            if second_to_last_node:
                if _is_fork_node(pred_node):
                    # add the entire "fork" to the region
                    for succ in graph.successors(pred_node):
                        region.add_edge(pred_node, succ)
                elif graph.out_degree[pred_node] != 1:
                    # the predecessor has more than one successor, and it's not a fork node
                    break

                if graph.in_degree[pred_node] == 1:
                    # continue search
                    pass
                else:
                    region.add_edge(pred_node, region_head)
                    traversed.add(pred_node)
                    region_head = pred_node
                    break
            elif not second_to_last_node and graph.out_degree[pred_node] != 1:
                break

            region.add_edge(pred_node, region_head)
            traversed.add(pred_node)
            region_head = pred_node

        return region, region_head

    @staticmethod
    def _is_indirect_jump_ailblock(block: Block) -> bool:
        if block.statements and isinstance(block.statements[-1], Jump):
            last_stmt = block.statements[-1]
            if not isinstance(last_stmt.target, Const):
                # it's an indirect jump (assuming the AIL block is properly optimized)
                return True
        return False

    @staticmethod
    def _is_single_return_stmt_region(region: networkx.DiGraph) -> bool:
        """
        Checks weather the provided region contains only one return statement. This stmt
        can be connected by many jumps, but none can be conditional. A valid case is:
        [Jmp] -> [Jmp] -> [Ret]
        """
        valid_stmt_types = (Return, Jump, Label)
        for node in region.nodes():
            if isinstance(node, Block):
                for stmt in node.statements:
                    if not isinstance(stmt, valid_stmt_types):
                        return False
        return True

    @staticmethod
    def _find_block_sets_in_all_regions(top_region: GraphRegion):
        def _unpack_block_type_to_addrs(node):
            if isinstance(node, Block):
                return {node.addr}
            if isinstance(node, MultiNode):
                return {n.addr for n in node.nodes}
            if isinstance(node, ConditionNode):
                return _unpack_block_type_to_addrs(node.true_node) | _unpack_block_type_to_addrs(node.false_node)
            return set()

        def _unpack_region_to_block_addrs(region: GraphRegion):
            region_addrs = set()
            for node in region.graph.nodes:
                if isinstance(node, (Block, MultiNode, ConditionNode)):
                    region_addrs |= _unpack_block_type_to_addrs(node)
                elif isinstance(node, GraphRegion):
                    region_addrs |= _unpack_region_to_block_addrs(node)

            return region_addrs

        def _unpack_every_region(region: GraphRegion, addrs_by_region: dict):
            addrs_by_region[region] = set()
            for node in region.graph.nodes:
                if isinstance(node, Block):
                    addrs_by_region[region].add(node.addr)
                elif isinstance(node, MultiNode):
                    for _node in node.nodes:
                        addrs_by_region[region].add(_node.addr)
                elif isinstance(node, ConditionNode):
                    addrs_by_region[region] |= _unpack_block_type_to_addrs(node.true_node)
                    addrs_by_region[region] |= _unpack_block_type_to_addrs(node.false_node)
                else:
                    addrs_by_region[region] |= _unpack_region_to_block_addrs(node)
                    _unpack_every_region(node, addrs_by_region)

        all_region_block_sets = {}
        _unpack_every_region(top_region, all_region_block_sets)
        return all_region_block_sets

    @staticmethod
    def _fix_copied_node_labels(block: Block):
        for i in range(len(block.statements)):  # pylint:disable=consider-using-enumerate
            stmt = block.statements[i]
            if isinstance(stmt, Label):
                # fix the default name by suffixing it with the new block ID
                new_name = stmt.name or f"Label_{stmt.tags['ins_addr']:x}"
                if "block_idx" in stmt.tags:
                    suffix = f"__{stmt.tags['block_idx']}"
                    new_name = new_name.removesuffix(suffix)
                else:
                    new_name = stmt.name
                new_name += f"__{block.idx}"

                block.statements[i] = Label(stmt.idx, new_name, **(stmt.tags | {"block_idx": block.idx}))

    @staticmethod
    def unwrap_conv(expr):
        return ReturnDuplicatorBase.unwrap_conv(expr.operand) if isinstance(expr, ailment.expression.Convert) else expr



from __future__ import annotations
import angr


class error(angr.SimProcedure):
    # pylint:disable=arguments-differ,missing-class-docstring

    NO_RET = None
    DYNAMIC_RET = True

    def run(self, status, errnum, fmt):  # pylint:disable=unused-argument
        fd = self.state.posix.get_fd(1)
        fprintf = angr.SIM_PROCEDURES["libc"]["fprintf"]
        self.inline_call(fprintf, fd, fmt)  # FIXME: This will not properly replace format strings

        if status.concrete and self.state.solver.eval(status) != 0:
            self.exit(status)

    def dynamic_returns(self, blocks, **kwargs) -> bool | None:
        # Execute those blocks with a blank state, and then dump the arguments
        blank_state = angr.SimState(
            project=self.project,
            mode="fastpath",
            cle_memory_backer=self.project.loader.memory,
            add_options={
                angr.options.SYMBOL_FILL_UNCONSTRAINED_MEMORY,
                angr.options.SYMBOL_FILL_UNCONSTRAINED_REGISTERS,
            },
        )
        # Execute each block
        state = blank_state
        for idx, b in enumerate(blocks):
            successors = self.project.factory.default_engine.process(state, irsb=b, force_addr=b.addr)
            if successors.all_successors:
                state = successors.all_successors[0]
                # if it was a call before reaching the last block, pick the one with Ijk_FakeRet
                if (
                    idx < len(blocks) - 1
                    and b.jumpkind
                    and (b.jumpkind == "Ijk_Call" or b.jumpkind.startswith("Ijk_Sys"))
                ):
                    fakerets = [succ for succ in successors.all_successors if succ.history.jumpkind == "Ijk_FakeRet"]
                    if fakerets:
                        state = fakerets[0]
            else:
                break

        # take a look at the first argument (status)
        cc = angr.default_cc(
            self.arch.name, platform=self.project.simos.name if self.project.simos is not None else None
        )(self.arch)
        ty = angr.sim_type.parse_signature("void x(int, int, char*)").with_arch(self.arch)
        args = cc.get_args(state, ty)
        # Return None if we can't determine the status
        if not args[0].concrete:
            return None
        return state.solver.eval(args[0]) == 0



from __future__ import annotations
from angr.ailment.expression import Load, UnaryOp

from .base import PeepholeOptimizationExprBase


class RemoveRedundantDerefs(PeepholeOptimizationExprBase):
    """
    Remove redundant dereferences (e.g. *(&v))
    """

    __slots__ = ()

    NAME = "Remove redundant dereferences"
    expr_classes = (Load,)

    def optimize(self, expr: Load, **kwargs):
        if isinstance(expr.addr, UnaryOp) and expr.addr.op == "Reference" and expr.bits == expr.addr.operand.bits:
            # *(&v) ==> v
            return expr.addr.operand
        return None



from __future__ import annotations
import angr


class vsnprintf(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, str_ptr, size, fmt, ap):  # pylint:disable=unused-argument
        # This function returns
        # Add another exit to the retn_addr that is at the top of the stack now

        if self.state.solver.eval(size) == 0:
            return 0

        self.state.memory.store(str_ptr, b"\x00")

        return 1



# pylint: disable=unused-argument
from __future__ import annotations


class Top:
    __slots__ = ("size",)

    def __init__(self, size):
        self.size = size

    @property
    def bits(self):
        return self.size * 8

    def __add__(self, other):
        return self

    def __and__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __rsub__(self, other):
        return self

    def __radd__(self, other):
        return self

    def __ror__(self, other):
        return self

    def __rand__(self, other):
        return self

    def __mul__(self, other):
        return self

    def __rmul__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __rfloordiv__(self, other):
        return self

    def __invert__(self):
        return self

    def __rshift__(self, other):
        return self

    def __rlshift__(self, other):
        return self

    def __rrshift__(self, other):
        return self

    def __lshift__(self, other):
        return self

    def __xor__(self, other):
        return self

    def __rxor__(self, other):
        return self

    def __or__(self, other):
        return self

    def __repr__(self):
        return "TOP"

    def __eq__(self, other):
        return Top(1)

    def __le__(self, other):
        return Top(1)

    def __lt__(self, other):
        return Top(1)

    def __gt__(self, other):
        return Top(1)

    def __ge__(self, other):
        return Top(1)

    def __neg__(self):
        return self

    def __hash__(self):
        return hash((Top, self.size))


class Bottom:
    def __add__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __and__(self, other):
        return self

    def __rshift__(self, other):
        return self

    def __repr__(self):
        return "BOTTOM"

    def __eq__(self, other):
        return type(other) is Bottom

    def __hash__(self):
        return hash(Bottom)



# pylint:disable=too-many-boolean-expressions
from __future__ import annotations
from collections import defaultdict
import logging

import angr.ailment as ailment

from angr.utils.bits import s2u
from angr.analyses.decompiler.stack_item import StackItem, StackItemType
from .optimization_pass import OptimizationPass, OptimizationPassStage

_l = logging.getLogger(name=__name__)


class StackCanarySimplifier(OptimizationPass):
    """
    Removes stack canary checks from decompilation results.
    """

    ARCHES = [
        "X86",
        "AMD64",
    ]  # TODO: fs is x86 only. Figure out how stack canary is loaded in other architectures
    PLATFORMS = ["cgc", "linux"]
    STAGE = OptimizationPassStage.AFTER_GLOBAL_SIMPLIFICATION
    NAME = "Simplify stack canaries"
    DESCRIPTION = __doc__.strip()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.analyze()

    def _check(self):
        # Check the first block and see if there is any statement reading data from fs:0x28h
        init_stmt = self._find_canary_init_stmt()

        return init_stmt is not None, {"init_stmt": init_stmt}

    def _analyze(self, cache=None):
        init_stmt = None
        if cache is not None:
            init_stmt = cache.get("init_stmt", None)

        if init_stmt is None:
            init_stmt = self._find_canary_init_stmt()

        if init_stmt is None:
            return

        # Look for the statement that loads back canary value from the stack
        first_block, stmt_idx = init_stmt
        canary_init_stmt = first_block.statements[stmt_idx]
        # where is the stack canary stored?
        if not (isinstance(canary_init_stmt.dst, ailment.Expr.VirtualVariable) and canary_init_stmt.dst.was_stack):
            _l.debug(
                "Unsupported canary storing location %s. Expects a stack VirtualVariable.",
                canary_init_stmt.addr,
            )
            return

        store_offset = canary_init_stmt.dst.stack_offset
        if not isinstance(store_offset, int):
            _l.debug("Unsupported canary storing offset %s. Expects an int.", store_offset)

        # The function should have at least one end point with an if-else statement
        # Find all nodes with 0 out-degrees
        all_endpoint_addrs = [node.addr for node in self._func.graph.nodes() if self._func.graph.out_degree(node) == 0]

        # Before node duplication, each pair of canary-check-success and canary-check-failure nodes have a common
        # predecessor.
        # map endpoint addrs to their common predecessors
        pred_addr_to_endpoint_addrs: dict[int, set[int]] = defaultdict(set)
        for node_addr in all_endpoint_addrs:
            preds = self._func.graph.predecessors(self._func.get_node(node_addr))
            for pred in preds:
                pred_addr_to_endpoint_addrs[pred.addr].add(node_addr)

        found_endpoints = False
        for pred_addr in pred_addr_to_endpoint_addrs:
            endpoint_addrs = pred_addr_to_endpoint_addrs[pred_addr]

            if len(endpoint_addrs) != 2:
                # we expect there to be only two nodes: one for canary-check-success, and the other for
                # canary-check-failure. if not, we check the next predecessor
                continue

            # because other optimization passes may duplicate nodes, we may have more than one node for each function
            # endpoint.
            endpoint_addrs_list = list(endpoint_addrs)
            endnodes_0 = list(self._get_blocks(endpoint_addrs_list[0]))
            endnodes_1 = list(self._get_blocks(endpoint_addrs_list[1]))

            if not endnodes_0 or not endnodes_1:
                _l.warning("Unexpected situation: endnodes_0 or endnodes_1 is empty.")
                continue

            # One of the end nodes calls __stack_chk_fail
            stack_chk_fail_callers = None
            other_nodes = None
            for endnodes, o in [(endnodes_0, endnodes_1), (endnodes_1, endnodes_0)]:
                if self._calls_stack_chk_fail(endnodes[0]):
                    stack_chk_fail_callers = endnodes
                    other_nodes = o
                    break
            else:
                _l.debug("Cannot find the node that calls __stack_chk_fail().")
                continue

            # Match stack_chk_fail_caller, ret_node, and predecessor
            nodes_to_process = []
            for stack_chk_fail_caller in stack_chk_fail_callers:
                all_preds = set(self._graph.predecessors(stack_chk_fail_caller))
                preds_for_other_nodes = set()
                for o in other_nodes:
                    preds_for_other_nodes |= set(self._graph.predecessors(o))
                preds = list(all_preds.intersection(preds_for_other_nodes))
                if len(preds) != 1:
                    _l.debug("Expect 1 predecessor. Found %d.", len(preds))
                    continue
                pred = preds[0]

                # More sanity checks
                if len(pred.statements) < 1:
                    _l.debug("The predecessor node is empty.")
                    continue

                # Check the last statement
                if not isinstance(pred.statements[-1], ailment.Stmt.ConditionalJump):
                    _l.debug("The predecessor does not end with a conditional jump.")
                    continue

                # Find the statement that computes real canary value xor stored canary value
                canary_check_stmt_idx = self._find_canary_comparison_stmt(pred, store_offset)
                if canary_check_stmt_idx is None:
                    _l.debug("Cannot find the canary check statement in the predecessor.")
                    continue

                succs = list(self._graph.successors(pred))
                if len(succs) != 2:
                    _l.debug("Expect 2 successors. Found %d.", len(succs))
                    continue
                ret_node = succs[1] if stack_chk_fail_caller is succs[0] else succs[0]
                nodes_to_process.append((pred, canary_check_stmt_idx, stack_chk_fail_caller, ret_node))

            # Awesome. Now patch this function.
            for pred, _canary_check_stmt_idx, stack_chk_fail_caller, ret_node in nodes_to_process:
                # Patch the pred so that it jumps to the one that is not stack_chk_fail_caller
                pred_copy = pred.copy()
                pred_copy.statements[-1] = ailment.Stmt.Jump(
                    len(pred_copy.statements) - 1,
                    ailment.Expr.Const(None, None, ret_node.addr, self.project.arch.bits),
                    ins_addr=pred_copy.statements[-1].tags["ins_addr"],
                )

                self._graph.remove_edge(pred, stack_chk_fail_caller)
                self._update_block(pred, pred_copy)

                if self._graph.in_degree[stack_chk_fail_caller] == 0:
                    # Remove the block that calls stack_chk_fail_caller
                    self._remove_block(stack_chk_fail_caller)

                found_endpoints = True

        if found_endpoints:
            # Remove the statement that loads the stack canary from fs
            first_block_copy = first_block.copy()
            first_block_copy.statements.pop(stmt_idx)
            self._update_block(first_block, first_block_copy)

            # update stack_items
            self.stack_items[store_offset] = StackItem(
                store_offset, canary_init_stmt.dst.size, "canary", StackItemType.STACK_CANARY
            )

        # Done!

    def _find_canary_init_stmt(self):
        block_addr = self._func.addr
        traversed = set()

        while True:
            traversed.add(block_addr)
            first_block = next(self._get_blocks(block_addr), None)
            if first_block is None:
                break

            for idx, stmt in enumerate(first_block.statements):
                if (
                    isinstance(stmt, ailment.Stmt.Assignment)
                    and isinstance(stmt.dst, ailment.Expr.VirtualVariable)
                    and stmt.dst.was_stack
                    and isinstance(stmt.dst.stack_offset, int)
                    and isinstance(stmt.src, ailment.Expr.Load)
                    and self._is_add(stmt.src.addr)
                ):
                    # Check addr: must be fs+0x28
                    op0, op1 = stmt.src.addr.operands
                    if isinstance(op1, ailment.Expr.VirtualVariable) and op1.was_reg:
                        op0, op1 = op1, op0
                    if (
                        isinstance(op0, ailment.Expr.VirtualVariable)
                        and op0.was_reg
                        and isinstance(op1, ailment.Expr.Const)
                        and op0.reg_offset == self.project.arch.get_register_offset("fs")
                        and op1.value == 0x28
                    ):
                        return first_block, idx

            succs = list(self._graph.successors(first_block))
            if len(succs) == 1:
                block_addr = succs[0].addr
                if block_addr not in traversed:
                    continue
            break

        return None

    def _find_canary_comparison_stmt(self, block, canary_value_stack_offset):
        for idx, stmt in enumerate(block.statements):
            if isinstance(stmt, ailment.Stmt.ConditionalJump):
                if isinstance(stmt.condition, ailment.Expr.UnaryOp) and stmt.condition.op == "Not":
                    # !(s_10 ^ fs:0x28 == 0)
                    negated = True
                    condition = stmt.condition.operand
                else:
                    negated = False
                    condition = stmt.condition
                if isinstance(condition, ailment.Expr.BinaryOp) and (
                    (not negated and condition.op == "CmpEQ") or (negated and condition.op == "CmpNE")
                ):
                    pass
                else:
                    continue

                expr0, expr1 = condition.operands
                if isinstance(expr0, ailment.Expr.BinaryOp) and expr0.op == "Xor":
                    # a ^ b
                    op0, op1 = expr0.operands
                    if isinstance(op0, ailment.Expr.VirtualVariable) and op0.was_reg:
                        # maybe op0 holds the value of another stack variable, like the following:
                        #
                        # 00 | 0x404e75 | LABEL_404e75:
                        # 01 | 0x404e75 | vvar_62{reg 16} = vvar_79{stack -64}
                        # 02 | 0x404e7a | vvar_66{reg 16} = (vvar_62{reg 16} ^ Load(addr=(0x28<64> + vvar_31{reg 208}),
                        #                 size=8, endness=Iend_LE))
                        # 03 | 0x404e83 | if (((vvar_62{reg 16} ^ Load(addr=(0x28<64> + vvar_31{reg 208}), size=8,
                        #                 endness=Iend_LE)) == 0x0<64>)) { Goto ... } else { Goto ... }
                        op0_v = self._get_vvar_value(block, op0.varid)
                        if isinstance(op0_v, ailment.Expr.VirtualVariable) and op0_v.was_stack:
                            op0 = op0_v

                    if not (
                        (
                            self._is_stack_canary_load_expr(op0, self.project.arch.bits, canary_value_stack_offset)
                            and self._is_random_number_load_expr(op1, self.project.arch.get_register_offset("fs"))
                        )
                        or (
                            self._is_stack_canary_load_expr(op1, self.project.arch.bits, canary_value_stack_offset)
                            and self._is_random_number_load_expr(op0, self.project.arch.get_register_offset("fs"))
                        )
                    ):
                        continue
                elif (
                    isinstance(expr0, ailment.Expr.Load)
                    and isinstance(expr1, ailment.Expr.Load)
                    and condition.op == "CmpEQ"
                ):
                    # a == b
                    if not (
                        (
                            self._is_stack_canary_load_expr(expr0, self.project.arch.bits, canary_value_stack_offset)
                            and self._is_random_number_load_expr(expr1, self.project.arch.get_register_offset("fs"))
                        )
                        or (
                            self._is_stack_canary_load_expr(expr1, self.project.arch.bits, canary_value_stack_offset)
                            and self._is_random_number_load_expr(expr0, self.project.arch.get_register_offset("fs"))
                        )
                    ):
                        continue

                # Found it
                return idx

        return None

    def _calls_stack_chk_fail(self, node):
        for stmt in node.statements:
            if isinstance(stmt, ailment.Stmt.SideEffectStatement) and isinstance(stmt.expr.target, ailment.Expr.Const):
                const_target = stmt.expr.target.value
                if const_target in self.kb.functions:
                    func = self.kb.functions.function(addr=const_target)
                    if func.name == "__stack_chk_fail":
                        return True

        return False

    @staticmethod
    def _is_stack_canary_load_expr(expr, bits: int, canary_value_stack_offset: int) -> bool:
        if not (
            isinstance(expr, ailment.Expr.VirtualVariable) and expr.was_stack and isinstance(expr.stack_offset, int)
        ):
            return False
        return s2u(expr.stack_offset, bits) == s2u(canary_value_stack_offset, bits)

    @staticmethod
    def _is_random_number_load_expr(expr, fs_reg_offset: int) -> bool:
        return (
            isinstance(expr, ailment.Expr.Load)
            and isinstance(expr.addr, ailment.Expr.BinaryOp)
            and expr.addr.op == "Add"
            and isinstance(expr.addr.operands[0], ailment.Expr.Const)
            and expr.addr.operands[0].value == 0x28
            and isinstance(expr.addr.operands[1], ailment.Expr.VirtualVariable)
            and expr.addr.operands[1].was_reg
            and expr.addr.operands[1].reg_offset == fs_reg_offset
        )

    @staticmethod
    def _get_vvar_value(block: ailment.Block, vvar_id: int) -> ailment.Expression | None:
        for stmt in block.statements:
            if (
                isinstance(stmt, ailment.Stmt.Assignment)
                and isinstance(stmt.dst, ailment.Expr.VirtualVariable)
                and stmt.dst.varid == vvar_id
            ):
                return stmt.src
        return None



from __future__ import annotations


class BlockID:
    """
    A context-sensitive key for a SimRun object.
    """

    def __init__(self, addr: int, callsite_tuples: tuple[int, ...] | None, jump_type: str):
        self.addr = addr
        self.callsite_tuples = callsite_tuples
        self.jump_type = jump_type

        self._hash = None

    def callsite_repr(self):
        if self.callsite_tuples is None:
            return "None"

        s = []

        def format_addr(addr):
            return "None" if addr is None else hex(addr)

        for i in range(0, len(self.callsite_tuples), 2):
            s.append("@".join(map(format_addr, self.callsite_tuples[i : i + 2])))  # pylint:disable=bad-builtin
        return " -> ".join(s)

    def __repr__(self):
        return f"<BlockID {self.addr:#08x} ({self.callsite_repr()}) % {self.jump_type}>"

    def __hash__(self):
        if self._hash is None:
            self._hash = hash((self.callsite_tuples, self.addr, self.jump_type))
        return self._hash

    def __eq__(self, other):
        return (
            isinstance(other, BlockID)
            and self.addr == other.addr
            and self.callsite_tuples == other.callsite_tuples
            and self.jump_type == other.jump_type
        )

    def __ne__(self, other):
        return not self == other

    @staticmethod
    def new(addr, callstack_suffix, jumpkind):
        if jumpkind.startswith("Ijk_Sys") or jumpkind == "syscall":
            jump_type = "syscall"
        elif jumpkind in ("Ijk_Exit", "exit"):
            jump_type = "exit"
        else:
            jump_type = "normal"
        return BlockID(addr, callstack_suffix, jump_type)

    @property
    def func_addr(self):
        if self.callsite_tuples:
            return self.callsite_tuples[-1]
        return None
